#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <octave/oct.h>
#include <octave/CColVector.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>
#include <octave/dbleSVD.h>
#include "channelflow/flowfield.h"
#include "channelflow/diffops.h"
#include "channelflow/utilfuncs.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/periodicfunc.h"
#include "octutils.h"

// John Gibson, Wed Oct 10 16:10:46 EDT 2007

using namespace std;
using namespace channelflow;

typedef FieldSymmetry Symm;
template <class T> void push(const T& t, array<T>& arrayt);

Real meandissipation(const FlowField& u, Real T, Real Reynolds, const DNSFlags& dnsflags,
		     const TimeStep& dt, SolutionType solntype);

Real observable(FlowField& u);

void symmetrize(FlowField& u, bool s1, bool s2, bool force, ostream* os);

int main(int argc, char* argv[]) {


  string purpose("(ALPHA VERSION!) continue invariant solution of plane Couette flow in Reynolds number");
  ArgList args(argc, argv, purpose);

  DNSFlags dnsflags;
  GMRESHookstepFlags srchflags;

  const bool restart    = args.getflag("-r",    "--restart",       "start from three previously computed solutions");
  const bool eqb        = args.getflag("-eqb",  "--equilibrium",   "search for equilibrium or relative equilibrium (trav wave)");
  const bool orb        = args.getflag("-orb",  "--periodicorbit", "search for periodic orbit or relative periodic orbit");
  const bool relative   = args.getflag("-rel",  "--relative",      "search over phase shifts ax,az, for relative orbit or eqb");

  const bool Recontinue = args.getflag("-contRe",  "--continueRe", "continue solution in Reynolds number");
  const bool Lxcontinue = args.getflag("-contLx",  "--continueLx", "continue solution in streamwise width Lx");
  const bool Lzcontinue = args.getflag("-contLz",  "--continueLz", "continue solution in spanwise width Lz");

  const Real T0         = args.getreal("-T",  "--maptime",   15.0,  "initial guess for orbit period or time of eqb/reqb map f^T(u)");
  const Real Re0        = args.getreal("-R",  "--Reynolds", 400.0,  "Reynolds number");
  const  Real s0        = args.getreal("-s0", "--s0" ,        0.0,  "start value for arclength (arbitrary)");
  /****/ Real ds        = args.getreal("-ds", "--ds" ,        0.01, "initial arclength increment for quadratic extrapolation");
  //const Real sfinal     = args.getreal("-sf",  "--sfinal",  1.0,  "final arclength (in normalized D,Re space)");

  const Real dsmin      = args.getreal("-dsmin", "--dsmin",   1e-04, "minimum arclength increment (in normalized D,Re space)");
  const Real dsmax      = args.getreal("-dsmax", "--dsmax",   0.05,  "maximum arclength increment (in normalized D,Re space)");
  const Real guesserrmin= args.getreal("-errmin", "--errmin", 1e-05, "minimum error for extrapolated guesses");
  const Real guesserrmax= args.getreal("-errmax", "--errmax", 1e-03, "maximum error for extrapolated guesses");

  const string sigmastr = args.getstr ("-sigma", "--sigma",      "", "file containing sigma of sigma f^T(u) - u = 0 (default == identity)");
  const string symmstr  = args.getstr ("-symms", "--symmetries", "", "file containing generators of isotropy group for symmetry-constrained search");

  dnsflags.baseflow     = PlaneCouette;
  dnsflags.timestepping = s2stepmethod(args.getstr("-ts", "--timestepping", "SBDF3", "time step algorithm: sbdf[1234]|cnab2|cnrk2|smrk2"));
  dnsflags.dealiasing   = DealiasXZ;
  dnsflags.nonlinearity = s2nonlmethod(args.getstr("-nl", "--nonlinearity", "Rotational", "nonlinearity method: rot|conv|skew|div|alt"));
  dnsflags.constraint   = PressureGradient;
  dnsflags.dPdx         = 0.0;
  dnsflags.verbosity    = Silent;
  if (symmstr.length() > 0) {
    SymmetryList symms(symmstr);
    cout << "Restricting flow to invariant subspace generated by symmetries" << endl;
    cout << symms << endl;
    dnsflags.symmetries = symms;
  }

  const bool vardt     = args.getflag("-vdt","--variableDt", "adjust dt to keep CFLmin<=CFL<CFLmax");
  const Real dtarg     = args.getreal("-dt","--timestep",  0.03125, "(initial) integration timestep");
  const Real dtmin     = args.getreal("-dtmin", "--dtmin", 0.001, "minimum timestep");
  const Real dtmax     = args.getreal("-dtmax", "--dtmax", 0.04,  "maximum timestep");
  const Real dTCFL     = args.getreal("-dTCFL", "--dTCFL", 1.00,  "check CFL # at this interval");
  const Real CFLmin    = args.getreal("-CFLmin", "--CFLmin", 0.40, "minimum CFL number");
  const Real CFLmax    = args.getreal("-CFLmax", "--CFLmax", 0.60, "maximum CFL number");

  //const Real epsMax     = args.getreal("-emax","--epsMax",      1e-10, "minimum acceptable residual for GMRESHookstep");
  srchflags.epsSearch = args.getreal("-es",  "--epsSearch",   1e-10, "stop search if L2Norm(s f^T(u) - u) < epsSearch");
  srchflags.epsKrylov = args.getreal("-ek",  "--epsKrylov",   1e-13, "min. condition # of Krylov vectors");
  srchflags.epsDu     = args.getreal("-edu", "--epsDuLinear", 1e-5,  "relative size of du to u in linearization");
  srchflags.epsDt     = args.getreal("-edt", "--epsDtLinear", 1e-7,  "size of dT in linearization of f^T about T");
  srchflags.epsGMRES  = args.getreal("-eg",  "--epsGMRES",    1e-3,  "stop GMRES iteration when Ax=b residual is < this");
  srchflags.epsGMRESf = args.getreal("-egf",  "--epsGMRESfinal",  0.05,  "accept final GMRES iterate if residual is < this");
  srchflags.centdiff  = args.getflag("-cd", "--centerdiff", "centered differencing to estimate differentials");

  srchflags.Nnewton   = args.getint( "-Nn", "--Nnewton",   10,    "max number of Newton steps ");
  srchflags.Ngmres    = args.getint( "-Ng", "--Ngmres",    60,    "max number of GMRES iterations per restart");
  srchflags.Nhook     = args.getint( "-Nh", "--Nhook",     20,    "max number of hookstep iterations per Newton");

  srchflags.delta     = args.getreal("-d",    "--delta",      0.01,  "initial radius of trust region");
  srchflags.deltaMin  = args.getreal("-dmin", "--deltaMin",   1e-12, "stop if radius of trust region gets this small");
  srchflags.deltaMax  = args.getreal("-dmax", "--deltaMax",   0.1,   "maximum radius of trust region");
  srchflags.deltaFuzz = args.getreal("-df",   "--deltaFuzz",  0.01,  "accept steps within (1+/-deltaFuzz)*delta");
  srchflags.lambdaMin = args.getreal("-lmin",   "--lambdaMin", 0.2,  "minimum delta shrink rate");
  srchflags.lambdaMax = args.getreal("-lmax",   "--lambdaMax", 1.5,  "maximum delta expansion rate");
  srchflags.lambdaRequiredReduction = 0.5; // when reducing delta, reduce by at least this factor.
  srchflags.improvReq = args.getreal("-irq",  "--improveReq", 1e-3,  "reduce delta and recompute hookstep if improvement "
				      " is worse than this fraction of what we'd expect from gradient");
  srchflags.improvOk  = args.getreal("-iok",  "--improveOk",   0.10, "accept step and keep same delta if improvement "
				      "is better than this fraction of quadratic model");
  srchflags.improvGood= args.getreal("-igd",  "--improveGood", 0.75, "accept step and increase delta if improvement "
				      "is better than this fraction of quadratic model");
  srchflags.improvAcc = args.getreal("-iac",  "--improveAcc",  0.10, "recompute hookstep with larger trust region if "
				      "improvement is within this fraction quadratic prediction.");

  srchflags.TscaleRel = args.getreal("-Tsc", "--Tscale",     20,    "scale time in hookstep: |T| = Ts*|u|/L2Norm(du/dt)");
  srchflags.dudtortho = args.getbool("-tc",  "--dudtConstr", true,  "require orthogonality of step to dudt");
  srchflags.orthoscale= args.getreal("-os",  "--orthoScale", 1,     "rescale orthogonality constraint");
  srchflags.dudxortho = args.getbool("-xc", "--dudxConstraint", true,"require orthogonality of step to dudx and dudz");
  srchflags.Rscale    = args.getreal("-rs", "--relativeScale", 1, "scale relative-search variables by this factor");
  srchflags.saveall   = args.getflag("-sa", "--saveall", "save each newton step");

  const string outdir   = args.getpath("-o",  "--outdir", "./", "output directory");
  const string logfile  = args.getstr ("-log","--logfile", "findsoln.log",  "output log (filename or \"stdout\")");

  string uname;
  string solndir[3];
  //Real ds0 = 0.01;
  bool up  = false;

  if (!restart) {
    //ds0 = args.getreal("-ds0", "--ds0" ,  0.01, "arclength increment for perturbations of initial solution");
    up  = args.getflag("-up", "--upwards",      "continue upwards in continuation param (Re, Lx, or Lz)");
    uname = args.getstr(1, "<flowfield>",       "initial guess for Newton search");
  }
  else {
    solndir[2]  = args.getpath(3, "<solndir0>", "directory containing previous solution data");
    solndir[1]  = args.getpath(2, "<solndir1>", "ditto");
    solndir[0]  = args.getpath(1, "<solndir2>", "ditto");
  }
  args.check();

  if (!(eqb^orb)) {
    cerr << "Please choose either -eqb or -orb option to search for (relative) equilibrium or (relative) periodic orbit" << endl;
    exit(1);
  }
  args.save("./");

  if (eqb && !relative)
    srchflags.solntype = Equilibrium;
  else if (eqb && relative)
    srchflags.solntype = TravelingWave;
  else if (orb && !relative)
    srchflags.solntype = PeriodicOrbit;
  else
    srchflags.solntype = RelativePeriodicOrbit;

  if (int(Recontinue) + int(Lxcontinue) + int(Lzcontinue) != 1) {
    cerr << "Please choose one of -contRe, -contLx, -contLz options for continuation in Re, Lx, or Lz" << endl;
    exit(1);
  }
  bool Tnormalize = eqb;

  Symm sigma0;
  if (sigmastr.length() != 0)
    sigma0 = Symm(sigmastr);

  int W = 24;
  cout << setprecision(17);
  cout << "Working directory == " << pwd() << endl;
  cout << "Command-line args == ";
  for (int i=0; i<argc; ++i) cout << argv[i] << ' '; cout << endl;

  cout << "Re    == " << Re0 << endl;
  cout << "sigma == " << sigma0 << endl;
  cout << "T     == " << T0 << endl;
  cout << "DNSFlags == " << dnsflags << endl << endl;

  array<FlowField> u(3); // solutions
  array<Real>      T(3); // period
  array<Real>      R(3); // Reynolds
  array<Real>      D(3); // dissipation (avg for orbits)
  array<Real>     mu(3); // continuation parameter
  array<Real>    obs(3); // observable
  array<Real>    res(3); // residual
  array<Real>      s(3); // (mu,obs) arclengths
  array<Symm>  sigma(3); // symmetry
  Real CFL = 0.0;
  int fcount = 0;

  ofstream dos((outdir + "ReD.asc").c_str());
  const TimeStep dt(dtarg, dtmin, dtmax, dTCFL, CFLmin, CFLmax, vardt);
  //TimeStep dt(dt0);


  // Compute initial data points for extrapolation from perturbations of given solution
  if (!restart) {
    u[2] = FlowField(uname);
    u[1] = FlowField(uname);
    u[0] = FlowField(uname);

    T[2] = T0;
    T[1] = T0;
    T[0] = T0;

    const int musign = up ? 1 : -1; // Go up or down in Re, Lx, or Lz?

    if (Recontinue) {
      R[2] = Re0*(1-musign*ds);
      R[1] = Re0;
      R[0] = Re0*(1+musign*ds);

      mu[2] = R[2];
      mu[1] = R[1];
      mu[0] = R[0];
    }
    else {
      R[2] = Re0;
      R[1] = Re0;
      R[0] = Re0;

      if (Lxcontinue) {
	mu[2] = u[2].Lx() * (1-musign*ds);
	mu[1] = u[1].Lx();
	mu[0] = u[0].Lx() * (1+musign*ds);

	u[2].rescale(mu[2], u[2].Lz());
	u[0].rescale(mu[0], u[0].Lz());
      }
      else {
	mu[2] = u[2].Lz() * (1-musign*ds);
	mu[1] = u[1].Lz();
	mu[0] = u[0].Lz() * (1+musign*ds);

	u[2].rescale(u[2].Lx(), mu[2]);
	u[0].rescale(u[0].Lx(), mu[0]);
      }
    }
    sigma[2] = sigma0;
    sigma[1] = sigma0;
    sigma[0] = sigma0;

  }
  // Read in some solutions from previous computations
  else {
     for (int i=0; i<3; ++i) {
       u[i] = FlowField(solndir[i] + "ubest");
       load(T[i], solndir[i]  + "Tbest");
       load(R[i], solndir[i] + "Reynolds");
       sigma[i] = Symm(solndir[i] + "sigmabest");
     }
     if (sigma[0] != sigma[1] || sigma[1] != sigma[2])
       cferror("continuesoln error : initial symmetries should be equal for non-relative searchs");
     sigma0 = sigma[1];

    if (Recontinue) {
      mu[2] = R[2];
      mu[1] = R[1];
      mu[0] = R[0];
    }
    else if (Lxcontinue) {
      mu[2] = u[2].Lx();
      mu[1] = u[1].Lx();
      mu[0] = u[0].Lx();
    }
    else {
      mu[2] = u[2].Lz();
      mu[1] = u[1].Lz();
      mu[0] = u[0].Lz();
    }

    cout << "loaded the following data..." << endl;
    cout << setw(4) << "i" << setw(W) << "T" << setw(W) << "Reynolds" << setw(W) << "L2Norm(u)" << setw(W) << "sigma" << endl;
    for (int i=2; i>=0; --i)
      cout << setw(4) << i << setw(W) << T[i] << setw(W) << R[i] << setw(W) << L2Norm(u[i]) << setw(W) << sigma[i] << endl;

  }

  // Now find solutions for initial data.
  for (int i=2; i>=0; --i) {

    // Allow for more newton steps if we start with blind perturbation of Re.
    GMRESHookstepFlags srchflags0 = srchflags;
    if (!restart)
      srchflags0.Nnewton = 20;

    cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl;
    string searchdir = outdir + "restart" + i2s(i) + "/";
    mkdir(searchdir);
    ofstream searchlog((searchdir + "findsoln.log").c_str());
    srchflags0.logstream = &searchlog;
    srchflags0.outdir    = searchdir;

    for (int n=0; n<3; ++n) {
      string label = "initial field u[" + i2s(n) + "]";
      project(dnsflags.symmetries, u[n], label, cout);
      fixdivnoslip(u[n]);
    }
    cout << "computing solution for mu == " << mu[i] << " in dir " << searchdir << "..." << endl;

    res[i] = GMRESHookstep(u[i], T[i], sigma[i], R[i], srchflags0, dnsflags, dt);
    //symmetrize(u[i], s1symm, s2symm, s1s2force, &cout);

    D[i]   = meandissipation(u[i], T[i], R[i], dnsflags, dt, srchflags0.solntype);
    obs[i] = D[i];

    cout << "mu == " << mu[i] << "   obs == " << obs[i] << "   D == " << D[i] << "   residual == " << res[i] <<endl;

  }

  Real obs0 = obs[1];
  Real mu0  = mu[1];

  s[2] = s0 - sqrt(square((obs[2]-obs[1])/obs0) + square((mu[2]-mu[1])/mu0));
  s[1] = s0;
  s[0] = s0 + sqrt(square((obs[0]-obs[1])/obs0) + square((mu[0]-mu[1])/mu0));

  //Real ds = sqrt(square((obs[0]-obs[1])/obs0) + square((mu[0]-mu[1])/mu0));

  Real guesserr = 0;
  dos << setw(W) << "% mu"<< setw(W) << "D" << setw(W) << "s" << setw(W) << "guesserr" << setw(W) << "residual" << "directory" << endl;
  for (int i=2; i>=0; --i)
    dos << setw(W) << mu[i] << setw(W) << D[i] << setw(W) << s[i] << setw(W) << guesserr << setw(W) << res[i] << "restart" << i << endl;

  int Nsteps = 100;
  bool failing=false;

  for (int n=0; n<Nsteps; ++n) {

    cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl;
    cout << "Continuing previous solutions by pseudo-arclength quadratic extrapolation in s = (mu,obs)" << endl;
    cout << setw(8) << "i" << setw(W) << "s" << setw(W) << "mu" << setw(W) << "T" << setw(W) << "obs" << endl;
    for (int i=2; i>=0; --i)
      cout << setw(8) << i << setw(W) << s[i] << setw(W) << mu[i] << setw(W) << T[i] << setw(W) << obs[i] << endl;

    Real snew = s[0] + ds;
    FlowField unew = quadraticInterpolate(u,  s, snew);
    Real Tnew      = isconst(T)  ? T[0]   : quadraticInterpolate(T, s, snew);
    Real Rnew      = !Recontinue ? Re0    : quadraticInterpolate(R, s, snew);
    Symm sigmanew  = !relative   ? sigma0 : quadraticInterpolate(sigma, s, snew);

    project(dnsflags.symmetries, unew, "quadratic extrapolant guess unew", cout);
    fixdivnoslip(unew);

    Real munew;
    if (Recontinue)      munew = Rnew;
    else if (Lxcontinue) munew = unew.Lx();
    else                 munew = unew.Lz();

    cout << setw(8) << "guess" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << endl;

    cout << "calculating error of quadratic extrapolated guess" << endl;
    FlowField Guguess;
    G(unew, Tnew, sigmanew, Guguess, Rnew, dnsflags, dt, Tnormalize, fcount, CFL);
    Real guesserr = L2Norm(Guguess);
    cout << '\n' << setw(8) << "guess" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << "   guess err == " << guesserr << endl;

    // Adjust ds on model err = k O(ds^3) , with fudge factor 2 or 1/2
    const int maxdsadjustments = 4;
    int dsadjustcount=0;
    Real guesserrtarget = sqrt(guesserrmin*guesserrmax); // aim between error bounds
    if (guesserr < guesserrmin)
      while (!failing && guesserr < guesserrmin && ds < dsmax && dsadjustcount++ < maxdsadjustments) {
	cout << "guess err == " << guesserr << " < guesserrmin == " << guesserrmin << ",  increasing ds..." << endl;
	// try to hit guesserrmindouble error based on err = O(ds^3) with 10% fudge factor
	// err = k ds^3 =>  err1/err0 = (ds1/ds0)^3 => ds1 = ds0 * (err1/err0)^(1/3)
	ds *=  pow(guesserrtarget/guesserr, 0.33);
	if (ds > dsmax) {
	  ds = dsmax;
	  dsadjustcount = maxdsadjustments;
	}
	snew = s[0] + ds;
	unew = quadraticInterpolate(u,  s, snew);
	Tnew = isconst(T)  ? T[0] : quadraticInterpolate(T,  s, snew);
	Rnew = !Recontinue ? R[0] : quadraticInterpolate(mu, s, snew);
	sigmanew  = !relative ? sigma0 : quadraticInterpolate(sigma, s, snew);

	project(dnsflags.symmetries, unew, "quadratic extrapolant guess u", cout);
	fixdivnoslip(unew);

	if (Recontinue)      munew = Rnew;
	else if (Lxcontinue) munew = unew.Lx();
	else                 munew = unew.Lz();

	G(unew, Tnew, sigmanew, Guguess, Rnew, dnsflags, dt, Tnormalize, fcount, CFL);
	guesserr = L2Norm(Guguess);
	cout << '\n' << setw(8) << "guess" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << "   guess err == " << guesserr << endl;
      }
    else if (guesserr > guesserrmax)
      while (guesserr > guesserrmax && ds > dsmin && dsadjustcount++ < 4) {
	cout << "guess err == " << guesserr << " > guesserrmax == " << guesserrmax << ",  decreasing ds..." << endl;
	ds *= pow(guesserrtarget/guesserr, 0.33);
	if (ds < dsmin) {
	  ds = dsmin;
	  dsadjustcount = maxdsadjustments;
	}
	snew = s[0] + ds;
	unew = quadraticInterpolate(u,  s, snew);
	Tnew = isconst(T)  ? T[0] : quadraticInterpolate(T,  s, snew);
	Rnew = !Recontinue ? R[0] : quadraticInterpolate(mu, s, snew);
	sigmanew  = !relative ? sigma0 : quadraticInterpolate(sigma, s, snew);

	if (Recontinue)      munew = Rnew;
	else if (Lxcontinue) munew = unew.Lx();
	else                 munew = unew.Lz();

	project(dnsflags.symmetries, unew, "quadratic extrapolant guess unew", cout);
	fixdivnoslip(unew);

	G(unew, Tnew, sigmanew, Guguess, Rnew, dnsflags, dt, Tnormalize, fcount, CFL);
	guesserr = L2Norm(Guguess);
	cout << '\n' << setw(8) << "guess" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << "   guess err == " << guesserr << endl;
      }

    if (guesserr < guesserrmin)
      cout << "\nFound guess whose error is too small but not too big. Going ahead with it.\n\n";
    else if (guesserr > guesserrmax)
      cout << "\nFound guess whose error is too big but not too small. Going ahead with it.\n\n" << endl;
    else
      cout << "\nFound guess whose error meets Goldilocks principle.\n\n";

    string searchdir = outdir + "search-" + i2s(n) + "/";
    mkdir(searchdir);
    save(Rnew, searchdir + "Reynolds");
    save(snew, searchdir + "arclength");
    save(munew, searchdir + "mu");
    ofstream searchlog((searchdir + "findsoln.log").c_str());

    //dnsflags.verbosity  = Silent;
    srchflags.logstream = &searchlog;
    srchflags.outdir    = searchdir;

    cout << "Computing new solution in directory " << searchdir << "  ..." << endl;

    //dt = dt0;
    Real resnew = GMRESHookstep(unew, Tnew, sigmanew, Rnew, srchflags, dnsflags, dt);
    //symmetrize(unew, s1symm, s2symm, s1s2force, &cout);
    Real Dnew = meandissipation(unew, Tnew, Rnew, dnsflags, dt, srchflags.solntype);
    Real obsnew = Dnew;

    if (resnew < srchflags.epsSearch) {
      cout << "Found new solution." << endl;

      save(Dnew,  searchdir + "D");
      save(Rnew,  searchdir + "Reynolds");
      save(Tnew,  searchdir + "T");
      sigmanew.save(searchdir + "sigmabest");
      unew.save(searchdir + "ubest"); // resave u after symmetrization

      //save(Tnew,  searchdir + "T");
      //save(Rnew,  searchdir + "Reynolds");
      //sigmanew.save(searchdir + "sigmabest");
      //unew.save(searchdir + "ubest");

      ds = sqrt(square((obsnew-obs[0])/obs0) + square((munew-mu[0])/mu0));
      snew = s[0]+ds;

      push(unew, u);
      push(Tnew, T);
      push(Rnew, R);
      push(Dnew, D);
      push(munew, mu);
      push(obsnew, obs);
      push(resnew, res);
      push(snew, s);
      push(sigmanew, sigma);

      cout << setw(8) << "solution" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << "   residual == " << resnew << endl;

      dos << setw(W) << munew << setw(W) << Dnew << setw(W) << snew << setw(W) << guesserr << setw(W) << resnew << searchdir << endl;
      //if (ds < dstarget)
      //ds = lesser(1.5*ds, dstarget);
      //else if (ds > dstarg)
      //ds = Greater(0.67*ds, dstarget);
      failing = false;
    }
    else {
      cout << "Failed to find new solution. Halving ds and trying again" << endl;
      cout << setw(8) << "failure" << setw(W) << snew << setw(W) << munew << setw(W) << Tnew << "   residual == " << resnew << endl;
      ds *= 0.5;
      failing = true;
      string failuredir(outdir + "failures");
      mkdir(failuredir);
      rename(searchdir, failuredir + "/search-" + i2s(n));
    }
  }
}

template <class T> void push(const T& t, array<T>& arrayt) {
  int N = arrayt.length();
  //cout << "push: array length = " << arrayt.length() << endl;
  //cfpause();
  for (int n=N-1; n>0; --n) {
    //cout << "pushing a[" << n-1 << "] -> a[" << n << "]" << endl;
    arrayt[n] = arrayt[n-1];
    //cfpause();
  }
  //cout << "pushing arg -> a[0]" << endl;
  arrayt[0] = t;
  //cfpause();
}


Real meandissipation(const FlowField& uarg, Real T, Real Reynolds, const DNSFlags& dnsflags,
		     const TimeStep& dtarg, SolutionType solntype) {

  cout << "computing mean dissipation..." << flush;
  FlowField u(uarg);

  ChebyCoeff U(u.Ny(),u.a(),u.b(),Spectral);
  U[1] = 1;

  if (solntype == Equilibrium || solntype == TravelingWave) {
    u += U;
    cout << "done" << endl;
    return dissipation(u);
  }
  else {

    //DNSFlags dnsflags = dnsflagsarg;
    //dnsflags.verbosity = PrintTicks;

    int N = iround(T);
    const Real dT = T/N;
    TimeStep dt(dtarg.dt(), dtarg.dtmin(), dtarg.dtmax(), dT, dtarg.CFLmin(), dtarg.CFLmax(), dtarg.variable());
    u.optimizeFFTW(FFTW_PATIENT);

    const int Nx = u.Nx();
    const int Ny = u.Ny();
    const int Nz = u.Nz();
    const Real Lx = u.Lx();
    const Real Lz = u.Lz();
    const Real a = u.a();
    const Real b = u.b();

    FlowField q(Nx,Ny,Nz,1,Lx,Lz,a,b);

    DNS dns(u, 1/Reynolds, dt, dnsflags, 0.0);

    int count = 0;
    Real D = 0.0;

    // Don't close the orbit so as not to count last point twice
    for (Real t=0; t<=T-dT/2; t += dT) {

      u += U;
      D += dissipation(u);
      u -= U;
      ++count;

      dns.advance(u, q, dt.n());

      if (dt.adjust(dns.CFL()))
	dns.reset_dt(dt);
   }
    cout << "done" << endl;
    return D/count;
  }
}

Real observable(FlowField& u) {
  ChebyCoeff U(u.Ny(),u.a(),u.b(),Spectral);
  U[1] = 1;
  u += U;
  Real obs = dissipation(u);
  u -= U;
  return obs;
}

/*******************************************************************88
void symmetrize(FlowField& u, bool s1opt, bool s2opt, bool force, ostream* os) {

  cout << "entering symmetrize..." << endl;
  FieldSymmetry tau;
  FieldSymmetry s1( 1, 1,-1, 0.5, 0);
  FieldSymmetry s2(-1,-1, 1, 0.5, 0.5);
  FieldSymmetry s3 = s1*s2;

  Real unorm2 = L2Norm2(u);
  Real eps = 1e-13;
  int Nsteps = 10;
  bool damp = false;
  bool verbose = false;
  if (os) {
    *os << endl;
    // *os.setf(ios::left);
    *os << "Initial energy decomposition:" << endl;
    *os << "s1 symm: " << setw(12) << project(u, s1, 1)/unorm2;
    *os << "   anti: " << setw(12) << project(u, s1,-1)/unorm2 << endl;
    *os << "s2 symm: " << setw(12) << project(u, s2, 1)/unorm2;
    *os << "   anti: " << setw(12) << project(u, s2,-1)/unorm2 << endl;
    *os << "s3 symm: " << setw(12) << project(u, s3, 1)/unorm2;
    *os << "   anti: " << setw(12) << project(u, s3,-1)/unorm2 << endl;
    *os << endl;
    // *os.unsetf(ios::left);
    // *os << "\nL2Dist(u,u0) == " << L2Dist(u,u0) << endl;
  }

  FieldSymmetry t;
  if (s1opt) {
    t = optimizePhase(u, s1, Nsteps, eps, damp, verbose);
    u = t(u);
    tau *= t;
  }
  if (s2opt) {
    t = optimizePhase(u, s2, Nsteps, eps, damp, verbose);
    u = t(u);
    tau *= t;
  }

  if (os && (s1opt || s2opt)) {
    *os << endl;
    // *os.setf(ios::left);
    *os << "Optimized energy decomposition:" << endl;
    *os << "s1 symm: " << setw(12) << project(u, s1, 1)/unorm2;
    *os << "   anti: " << setw(12) << project(u, s1,-1)/unorm2 << endl;
    *os << "s2 symm: " << setw(12) << project(u, s2, 1)/unorm2;
    *os << "   anti: " << setw(12) << project(u, s2,-1)/unorm2 << endl;
    *os << "s3 symm: " << setw(12) << project(u, s3, 1)/unorm2;
    *os << "   anti: " << setw(12) << project(u, s3,-1)/unorm2 << endl;
    *os << endl;
    // *os.unsetf(ios::left);
    *os << "Post-translation s1     symm err == " << L2Dist(u,s1(u)) << endl;
  }
  if (force) {
    if (s1opt) {
      if (os) *os << "\nForcing s1 symmetry by projection." << endl;
      (u += s1(u)) *= 0.5;
    }
    if (s2opt) {
      if (os) *os << "Forcing s2  symmetry by projection." << endl;
      (u += s2(u)) *= 0.5;
    }

    unorm2 = L2Norm2(u);
    if (os) {
      *os << endl;
      *os << "Forced energy decomposition:" << endl;
      // *os.setf(ios::left);
      *os << "s1 symm: " << setw(12) << project(u, s1, 1)/unorm2;
      *os << "   anti: " << setw(12) << project(u, s1,-1)/unorm2 << endl;
      *os << "s2 symm: " << setw(12) << project(u, s2, 1)/unorm2;
      *os << "   anti: " << setw(12) << project(u, s2,-1)/unorm2 << endl;
      *os << "s3 symm: " << setw(12) << project(u, s3, 1)/unorm2;
      *os << "   anti: " << setw(12) << project(u, s3,-1)/unorm2 << endl;
      *os << endl;
    }
  }
  u.makeSpectral();
  u.zeroPaddedModes();
  cout << "exiting symmetrize..." << endl;
}
************************************************************/

/* continuesoln.cpp, channelflow-1.3, www.channelflow.org
 *
 * Copyright (C) 2001-2009  John F. Gibson
 *
 * gibson@cns.physics.gatech.edu
 * Center for Nonlinear Sciences, School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */
