#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <octave/oct.h>
#include <octave/CColVector.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>
#include <octave/dbleSVD.h>
#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/periodicfunc.h"
#include "octutils.h"


using namespace std;
using namespace channelflow;


// This program is an implementation of Divakar Viswanath's Newton-hookstep
// GMRES algorithm for computing invariant solutions of Navier-Stokes. The
// basic idea is to find solutions of sigma f^T(u) - u = 0 where f^T(u) is the
// time-T CFD integration of the fluid velocity field u and sigma is a symmetry
// operation. T and sigma can be held fixed or varied. Viswanath's algorithm
// combines a Newton-hookstep trust-region minimization of ||sigma f^T(u) - u||^2
// with iterative GMRES solution of the Newton-step equations.

// In this program, the details of GMRES algorithm are modeled on Lecture 35
// of Trefethen and Bau's "Numerical Linear Algebra". The details of the hookstep
// algorithm are based on section 6.4.2 of Dennis and Schnabel's "Numerical Methods
// for Unconstrained Optimization and Nonlinear Equations."
//
// See octutils.h for a detailed description of the algorithm, notation,
// and references.
//
// John Gibson, Wed Oct 10 16:10:46 EDT 2007


int main(int argc, char* argv[]) {

  string purpose("Newton-Krylov-hookstep search for (relative) equilibrium or periodic orbit of plane Couette flow");
  ArgList args(argc, argv, purpose);

  DNSFlags dnsflags;
  GMRESHookstepFlags srchflags;

  const bool eqb        = args.getflag("-eqb",  "--equilibrium",   "search for equilibrium or relative equilibrium (trav wave)");
  const bool orb        = args.getflag("-orb",  "--periodicorbit", "search for periodic orbit or relative periodic orbit");
  const bool relative   = args.getflag("-rel",  "--relative",      "search over phase shifts ax,az, for relative orbit or eqb");

  /***/ Real T          = args.getreal("-T", "--maptime",  10.0,  "initial guess for orbit period or time of eqb/reqb map f^T(u)");
  const Real Reynolds   = args.getreal("-R", "--Reynolds", 400.0, "Reynolds number");
  const string symmname = args.getstr ("-s", "--symmetry", "",    "file containing symmetry (default == identity)");

  dnsflags.baseflow     = PlaneCouette;
  dnsflags.timestepping = s2stepmethod(args.getstr("-ts", "--timestepping", "SBDF3", "time step algorithm: sbdf[1234]|cnab2|cnrk2|smrk2"));
  dnsflags.dealiasing   = DealiasXZ;
  dnsflags.nonlinearity = s2nonlmethod(args.getstr("-nl", "--nonlinearity", "Rotational", "nonlinearity method: rot|conv|skew|div|alt"));
  dnsflags.constraint   = PressureGradient;
  dnsflags.dPdx         = 0.0;
  dnsflags.verbosity    = Silent;

  const bool vardt     = args.getflag("-vdt","--variableDt", "adjust dt to keep CFLmin<=CFL<CFLmax");
  const Real dtarg     = args.getreal("-dt","--timestep",  0.03125, "(initial) integration timestep");
  const Real dtmin     = args.getreal("-dtmin", "--dtmin", 0.001, "minimum timestep");
  const Real dtmax     = args.getreal("-dtmax", "--dtmax", 0.04,  "maximum timestep");
  const Real dTCFL     = args.getreal("-dTCFL", "--dTCFL", 1.00,  "check CFL # at this interval");
  const Real CFLmin    = args.getreal("-CFLmin", "--CFLmin", 0.40, "minimum CFL number");
  const Real CFLmax    = args.getreal("-CFLmax", "--CFLmax", 0.60, "maximum CFL number");

  srchflags.epsSearch = args.getreal("-es",  "--epsSearch",   1e-13, "stop search if L2Norm(s f^T(u) - u) < epsEQB");
  srchflags.epsKrylov = args.getreal("-ek",  "--epsKrylov",   1e-14, "min. condition # of Krylov vectors");
  srchflags.epsDu     = args.getreal("-edu", "--epsDuLinear", 1e-7,  "relative size of du to u in linearization");
  srchflags.epsDt     = args.getreal("-edt", "--epsDtLinear", 1e-5,  "size of dT in linearization of f^T about T");
  srchflags.epsGMRES  = args.getreal("-eg",  "--epsGMRES",    1e-3,  "stop GMRES iteration when Ax=b residual is < this");
  srchflags.epsGMRESf = args.getreal("-egf",  "--epsGMRESfinal",  0.05,  "accept final GMRES iterate if residual is < this");
  srchflags.centdiff  = args.getflag("-cd", "--centerdiff", "centered differencing to estimate differentials");

  srchflags.Nnewton   = args.getint( "-Nn", "--Nnewton",   20,    "max number of Newton steps ");
  srchflags.Ngmres    = args.getint( "-Ng", "--Ngmres",    40,    "max number of GMRES iterations per restart");
  srchflags.Nhook     = args.getint( "-Nh", "--Nhook",     20,    "max number of hookstep iterations per Newton");

  srchflags.delta     = args.getreal("-d",    "--delta",      0.01,  "initial radius of trust region");
  srchflags.deltaMin  = args.getreal("-dmin", "--deltaMin",   1e-12, "stop if radius of trust region gets this small");
  srchflags.deltaMax  = args.getreal("-dmax", "--deltaMax",   0.1,   "maximum radius of trust region");
  srchflags.deltaFuzz = args.getreal("-df",   "--deltaFuzz",  0.01,  "accept steps within (1+/-deltaFuzz)*delta");
  srchflags.lambdaMin = args.getreal("-lmin",   "--lambdaMin", 0.2,  "minimum delta shrink rate");
  srchflags.lambdaMax = args.getreal("-lmax",   "--lambdaMax", 1.5,  "maximum delta expansion rate");
  srchflags.lambdaRequiredReduction = 0.5; // when reducing delta, reduce by at least this factor.
  srchflags.improvReq = args.getreal("-irq",  "--improveReq", 1e-3,  "reduce delta and recompute hookstep if improvement "
				      " is worse than this fraction of what we'd expect from gradient");
  srchflags.improvOk  = args.getreal("-iok",  "--improveOk",   0.10, "accept step and keep same delta if improvement "
				      "is better than this fraction of quadratic model");
  srchflags.improvGood= args.getreal("-igd",  "--improveGood", 0.75, "accept step and increase delta if improvement "
				      "is better than this fraction of quadratic model");
  srchflags.improvAcc = args.getreal("-iac",  "--improveAcc",  0.10, "recompute hookstep with larger trust region if "
				      "improvement is within this fraction quadratic prediction.");

  srchflags.TscaleRel = args.getreal("-Tsc", "--Tscale",     20,    "scale time in hookstep: |T| = Ts*|u|/L2Norm(du/dt)");
  srchflags.dudtortho = args.getbool("-tc",  "--dudtConstr", true,  "require orthogonality of step to dudt");
  srchflags.orthoscale= args.getreal("-os",  "--orthoScale", 1,     "rescale orthogonality constraint");
  srchflags.dudxortho = args.getbool("-xc", "--dudxConstraint", true,"require orthogonality of step to dudx and dudz");
  srchflags.Rscale    = args.getreal("-rs", "--relativeScale", 1, "scale relative-search variables by this factor");
  srchflags.outdir    = args.getpath("-o",  "--outdir", "./", "output directory");
  const string logfile  = args.getstr ("-log","--logfile", "findsoln.log",  "output log (filename or \"stdout\")");
  const string uname    = args.getstr(1, "<flowfield>", "initial guess for Newton search");

  args.check();
  if (!(eqb^orb)) {
    cerr << "Please choose either -eqb or -orb option to search for (relative) equilibrium or (relative) periodic orbit" << endl;
    exit(1);
  }
  args.save("./");

  if (eqb && !relative)
    srchflags.solntype = Equilibrium;
  else if (eqb && relative)
    srchflags.solntype = TravelingWave;
  else if (orb && !relative)
    srchflags.solntype = PeriodicOrbit;
  else
    srchflags.solntype = RelativePeriodicOrbit;

  TimeStep dt(dtarg, dtmin, dtmax, dTCFL, CFLmin, CFLmax, vardt);
  FieldSymmetry sigma;
  if (symmname.length() != 0)
    sigma = FieldSymmetry(symmname);

  ofstream logstream;
  if (logfile == "stdout" || logfile == "cout")
    srchflags.logstream = &cout;
  else {
    logstream.open((srchflags.outdir+logfile).c_str());
    srchflags.logstream = &logstream;
  }
  ostream& os = *(srchflags.logstream); // a short name for ease of use


  os << "Working directory == " << pwd() << endl;
  os << "Command-line args == ";
  for (int i=0; i<argc; ++i) cout << argv[i] << ' '; cout << endl;

  os << "Re    == " << Reynolds << endl;
  os << "sigma == " << sigma << endl;
  os << "T     == " << T << endl;
  os << "dt    == " << dt << endl;
  os << "DNSFlags == " << dnsflags << endl << endl;

  FlowField u(uname);

  Real residual = GMRESHookstep(u, T, sigma, Reynolds, srchflags, dnsflags, dt);

  os << "Final search residual is " << residual << endl;
  os << "Saving final values of u,T,sigma residual is " << residual << endl;
  u.save(srchflags.outdir + "usoln");
  sigma.save(srchflags.outdir + "sigmasoln");
  save(T, srchflags.outdir + "Tsoln");
  save(Reynolds, srchflags.outdir + "Resoln");

  fftw_cleanup();
}




/* findorbit.cpp, channelflow-1.3, www.channelflow.org
 *
 * Copyright (C) 2001-2009  John F. Gibson
 *
 * gibson@cns.physics.gatech.edu
 * Center for Nonlinear Sciences, School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */
