#include <sys/stat.h>
#include <fstream>
#include <iomanip>
#include <string>
#include "octutils.h"
#include "channelflow/utilfuncs.h"
#include "channelflow/mathdefs.h"
#include "channelflow/periodicfunc.h"

using namespace std;
//using namespace channelflow;

namespace channelflow {

//void save(Real T, const std::string& filebase) {
//string filename = appendSuffix(filebase, ".asc");
//ofstream os(filename.c_str());
//os << setprecision(17);
//os << T << endl;
//};

void save(const std::string& filebase, Real  T, const FieldSymmetry& tau) {
  string filename = appendSuffix(filebase, ".asc");
  ofstream os(filename.c_str());
  const int p = 16;  // digits of precision
  const int w = p+7; // width of field
  os << setprecision(p) << scientific;
  os << setw(w) << T << " %T\n";
  os << setw(w) << tau.s() << " %s\n";
  os << setw(w) << tau.sx() << " %sx\n";
  os << setw(w) << tau.sy() << " %sy\n";
  os << setw(w) << tau.sz() << " %sz\n";
  os << setw(w) << tau.ax() << " %ax\n";
  os << setw(w) << tau.az() << " %az\n";
};

void load(const std::string& filebase, Real& T, FieldSymmetry& tau) {
  string filename = appendSuffix(filebase, ".asc");
  ifstream is(filename.c_str());
  string junk;
  int su, sx, sy, sz;
  Real ax, az;
  bool error = false;
  is >> T >> junk;
  if (junk != "%T") error = true;
  is >> su >> junk;
  if (junk != "%s" || abs(su) != 1) error = true;
  is >> sx >> junk;
  if (junk != "%sx" || abs(sx) != 1) error = true;
  is >> sy >> junk;
  if (junk != "%sy" || abs(sy) != 1) error = true;
  is >> sz >> junk;
  if (junk != "%sz" || abs(sz) != 1) error = true;
  is >> ax >> junk;
  if (junk != "%ax" || abs(ax) > 0.5) error = true;
  is >> az >> junk;
  if (junk != "%az" || abs(az) > 0.5) error = true;

  if (error) {
    cout << "Error in loading T,su,sx,sy,sz,ax,az from file " << filename << endl;
    cout << "Exiting." << endl;
    exit(1);
  }
  tau = FieldSymmetry(sx,sy,sz,ax,az,su);
}

void plotfield(const FlowField& u_, const string& outdir, const string& label,
	       int xstride, int ystride, int zstride) {

  string preface = outdir;
  if (label.length() > 0)
    preface += label + "_";

  mkdir(outdir);

  FlowField& u = (FlowField&) u_;
  fieldstate xs = u.xzstate();
  fieldstate ys = u.ystate();

  const int Ny2 = (u.Ny()-1)/2;
  //Vector x = periodicpoints(Nx, u.Lx());
  //Vector y = chebypoints(Ny, u.a(), u.b());
  //Vector z = periodicpoints(Nz, u.Lz());
  //x.save(outdir + "x");
  //y.save(outdir + "y");
  //z.save(outdir + "z");

  u.makePhysical();
  u.saveSlice(0, 0, 0,   preface + "u_yz");
  u.saveSlice(0, 1, 0,   preface + "v_yz");
  u.saveSlice(0, 2, 0,   preface + "w_yz");
  u.saveSlice(1, 0, Ny2, preface + "u_xz");
  u.saveSlice(1, 1, Ny2, preface + "v_xz");
  u.saveSlice(1, 2, Ny2, preface + "w_xz");
  u.saveSlice(1, 0, 11, preface + "u_xz_11");
  u.saveSlice(1, 1, 11, preface + "v_xz_11");
  u.saveSlice(1, 2, 11, preface + "w_xz_11");
  u.saveSlice(1, 0, 23, preface + "u_xz_23");
  u.saveSlice(1, 1, 23, preface + "v_xz_23");
  u.saveSlice(1, 2, 23, preface + "w_xz_23");
  u.saveSlice(2, 0, 0,   preface + "u_xy");
  u.saveSlice(2, 1, 0,   preface + "v_xy");
  u.saveSlice(2, 2, 0,   preface + "w_xy");
  u.makeSpectral();


  //FlowField vort = curl(u);
  //vort.makePhysical();
  //vort.saveSlice(0, 0, 0, preface + "vort_yz");

  //FlowField drag = ydiff(u);
  //drag.makePhysical();
  //drag.saveSlice(1, 0, Ny-1, preface + "drag_xz");

  //ChebyCoeff U = Re(u.profile(0,0,0));
  //U.makePhysical();
  //U.save("U");

  u.makeState(xs,ys);
}

// Produce plots of various slices, etc.
void plotxavg(const FlowField& u_, const std::string& outdir,
	      const std::string& label) {

  FlowField& u = (FlowField&) u_;
  fieldstate xs = u.xzstate();
  fieldstate ys = u.ystate();
  u.makeSpectral();

  FlowField uavg(4, u.Ny(), u.Nz(), u.Nd(), u.Lx(), u.Lz(), u.a(), u.b(), Spectral, Spectral);
  for (int i=0; i<u.Nd(); ++i)
    for (int my=0; my<u.My(); ++my)
      for (int mz=0; mz<u.Mz(); ++mz)
	uavg.cmplx(0,my,mz,i) = u.cmplx(0,my,mz,i);

  mkdir(outdir);
  string preface = outdir;
  if (label.length() > 0)
    preface += label + "_";

  uavg.makePhysical();
  uavg.saveSlice(0, 0, 0,   preface + "u_yz_xavg");
  uavg.saveSlice(0, 1, 0,   preface + "v_yz_xavg");
  uavg.saveSlice(0, 2, 0,   preface + "w_yz_xavg");

  u.makeState(xs,ys);
}


void plotspectra(const FlowField& u_, const string& outdir, const string& label,
		 bool showpadding) {

  FlowField& u = (FlowField&) u_;
  fieldstate xs = u.xzstate();
  fieldstate ys = u.ystate();

  string preface = outdir;
  if (label.length() > 0)
    preface += label + "_";

  mkdir(outdir);
  u.makeSpectral();
  u.saveSpectrum(preface + "xzspec", true, showpadding);

  u.makePhysical_y();
  ofstream ucs((preface + "yspec.asc").c_str());

  ucs << "% sum of abs vals of Cheby coeffs of components u_{kx,kz}(y) " << endl;

  ChebyTransform trans(u.Ny());

  const int kxmax = lesser(u.kxmax(), 8);
  const int kzmax = lesser(u.kzmax(), 8);
  const int dk = 1;

  for (int kx=0; kx<=kxmax; kx += dk) {
    //for (int kz=0; kz<=kzmax; kz += dk) {
    int kz=lesser(kx,kzmax);
    BasisFunc prof = u.profile(u.mx(kx),u.mz(kz));
      prof.makeSpectral(trans);
      ucs << "% kx,kz == " << kx << ", " << kz << endl;
      for (int ny=0; ny<prof.Ny(); ++ny) {
	Real sum = 0.0;
	for (int i=0; i<prof.Nd(); ++i)
	  sum += abs(prof[i][ny]);
	ucs << sum << ' ';
      }
      ucs << '\n';
  }

  u.makeState(xs,ys);

}

/******************8
string pathfix(const string& path) {
  string rtn = path;
  if (rtn.length() > 0 && rtn[rtn.length()-1] != '/')
    rtn += "/";
  return rtn;
}

void mkdir(const string& dirname) {
  ::mkdir(dirname.c_str(), S_IRUSR | S_IWUSR | S_IXUSR);
}

Real PuFraction(const FlowField& u, const FieldSymmetry& s, int sign) {
  FlowField Pu = s(u);
  if (sign == 1)
    Pu += u;
  else
    Pu -= u;
  return 0.25*L2Norm2(Pu)/L2Norm2(u); // 0.25 is square of 1/2 from Pu = 1/2 (u+s(u))
}
**********************/

// Return a translation that maximizes the s-symmetry of u.
// (i.e. return tau // that minimizes L2Norm(s(tau(u)) - tau(u))).
FieldSymmetry optimizePhase(const FlowField& u, const FieldSymmetry& s,
			    int Nsteps, Real residual, Real damp,
			    bool verbose, Real x0, Real z0) {

  FieldSymmetry tau;
  if (s.sx()==1 && s.sz()==1)
    ;
  else if ((s.sx() == -1) ^ (s.sz() == -1)) {

    // Newton search for x-translation that maximizes shift-rotate symmetry
    if (verbose)
      cout << "\noptimizePhase: Newton search on translation" <<endl;

    int i   = (s.sx() == -1) ? 0 : 2; // are we optimizing on x or z?
    Real L  = (s.sx() == -1) ? u.Lx() : u.Lz();
    array<Real> x(3);
    x[0] = x0;
    x[2] = z0;

    FlowField tu, stu, tux, stux, tuxx, stuxx;

    for (int n=0; n<Nsteps; ++n) {
      tau = FieldSymmetry(x[0]/u.Lx(), x[2]/u.Lz());

      tu  = tau(u);
      stu = s(tu);
      tux   = diff(tu, i, 1);
      stux  = diff(stu, i, 1);
      tuxx  = diff(tux, i, 1);
      stuxx = diff(stux, i, 1);

      // g(x) is df/dx where f == (L2Norm(stu - tu)) and t = tau(x/Lx,0)
      Real g    =  L2IP(stux, tu) - L2IP(stu, tux);
      Real dgdx = 2*L2IP(stux, tux) - L2IP(stuxx, tu) - L2IP(stu, tuxx);

      if (verbose) {
	Real f = L2Dist(stu, tu);
	cout << endl;
	cout << "Newton step n  == " << n << endl;
	cout << "tau.ax         == " << tau.ax() << endl;
	cout << "tau.az         == " << tau.az() << endl;
	cout << "zeroable g     == " << g << endl;
	cout << "L2Norm(stu-tu) == " << f << endl;
      }
      if (abs(g) < residual && verbose) {
	cout << "\noptimizePhase: Newton search converged." << endl;
	break;
      }

      Real dx = -g/dgdx;

      if (abs(damp*dx)/L > 0.5) {
	cout << "\nerror in optimizePhase(FlowField, FieldSymmetry) :" << endl;
	cout << "Newton search went beyond bounds. " << endl;
	cout << "     g == " << g << endl;
	cout << " dg/dx == " << dgdx << endl;
	cout << "    dx == " << damp*dx << endl;
	cout << "  L/2  == " << 0.5*L  << endl;
	cout << "Jostle by L/10 and try again." << endl;
	x[i] += L/10;
      }
      else
	x[i] += damp*dx;

    }
  }
  else {
    cout << "\nerror in optimizePhase(FlowField, FieldSymmetry) :" << endl;
    cout << "2d symmetrization is not yet implemented. Returning identity." << endl;
  }

  if (verbose) {
    FlowField tu = tau(u);
    FlowField stu = s(tu);
    cout << "\noptimal translation == " << tau << endl;
    cout << "Post-translation symmetry error == " << L2Dist(stu, tu) << endl;
  }
  return tau;
}

string t2s(Real t, bool inttime) {
  char buffer[12];
  if (inttime) {
    int it = iround(t);
    sprintf(buffer, "%d", it);
  }
  else
    sprintf(buffer, "%.3f", t);
  return string(buffer);
}

string fuzzyless(Real x, Real eps) {
  string rtn("false ");
  if (x < eps)
    rtn = "TRUE  ";
  else if (x < sqrt(eps))
    rtn = "APPROX";
  return rtn;
};

string pwd() {
  const int N=512;
  char* buff = new char[N];
  getcwd(buff, N);
  string rtn(buff);
  delete[] buff;
  return rtn;
}

string stub(const string& filename, const string& ext) {

  // Clip off path
  size_t s = filename.find_last_of("/");
  string f = (s == string::npos) ? filename : filename.substr(s+1, filename.length()-s);

  // Clip off extension
  size_t t = f.find(ext, f.length()-ext.length());
  string g = (t == string::npos) ? f : f.substr(0,t);

  return g;
}


// ======================================================================
// All Octave-dependent code goes within this ifdef
// #ifdef HAVE_OCTAVE

void setToZero(ColumnVector& x) {
  for (int i=0; i<x.length(); ++i)
    x(i) = 0.0;
}

void setToZero(Matrix& A) {
  for (int i=0; i<A.rows(); ++i)
    for (int j=0; j<A.cols(); ++j)
      A(i,j) = 0.0;
}

void save(const Matrix& A, const string& filebase) {
  string filename(filebase + ".asc");
  ofstream os(filename.c_str());
  os << setprecision(17);
  for (int i=0; i<A.rows(); ++i) {
    for (int j=0; j<A.cols(); ++j)
      os << A(i,j) << ' ';
    os << '\n';
  }
}

void print(const ColumnVector& x) {
  for (int i=0; i<x.length(); ++i)
    cout << x(i) << '\n';
}
void print(const Matrix& A) {
  for (int i=0; i<A.rows(); ++i) {
    for (int j=0; j<A.cols(); ++j)
      cout << A(i,j) << ' ';
    cout << '\n';
  }
}


void save(const ColumnVector& x, const string& filebase) {
  string filename(filebase + ".asc");
  ofstream os(filename.c_str());
  os << setprecision(17);
  for (int i=0; i<x.length(); ++i)
    os << x(i) << '\n';
}

void save(const ComplexColumnVector& x, const string& filebase) {
  string filename(filebase + ".asc");
  ofstream os(filename.c_str());
  os << setprecision(17);
  for (int i=0; i<x.length(); ++i)
    os << x(i).real() << ' ' << x(i).imag() << '\n';
}



Real L2Norm2(const ColumnVector& x) {
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += x(i)*x(i);
  return sum;
}

Real L2Dist2(const ColumnVector& x,  const ColumnVector& y) {
  assert(x.length() == y.length());
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += square(x(i)-y(i));
  return sum;
}
Real L2Norm(const ColumnVector& x) {
  return sqrt(L2Norm2(x));
}

Real L2Dist(const ColumnVector& x,  const ColumnVector& y) {
  return sqrt(L2Dist2(x,y));
}

Real V2Norm2(const FlowField& u) {
  ColumnVector v;
  field2vector(u,v);
  return L2Norm2(v);
}

Real V2Norm(const FlowField& u) {
  return sqrt(V2Norm2(u));
}

Real V2IP(const FlowField& u, const FlowField& v) {
  ColumnVector uV;
  ColumnVector vV;
  field2vector(u,uV);
  field2vector(v,vV);
  return (uV.transpose())*vV;
}

Real Norm2(const FlowField& u, BasisType norm) {
  if (norm == L2Basis)
    return L2Norm2(u);
  else
    return V2Norm2(u);
}

Real Norm(const FlowField& u, BasisType norm) {
  if (norm == L2Basis)
    return L2Norm(u);
  else
    return V2Norm(u);
}

Real IP(const FlowField& u, const FlowField& v, BasisType norm) {
  if (norm == L2Basis)
    return L2IP(u,v);
  else
    return V2IP(u,v);
}

void operator*=(ColumnVector& x, Real c) {
  for (int i=0; i<x.length(); ++i)
    x(i) *= c;
}

Real L2Norm2(const ComplexColumnVector& x) {
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += abs2(x(i));
  return sum;
}

Real L2Dist2(const ComplexColumnVector& x,  const ComplexColumnVector& y) {
  assert(x.length() == y.length());
  Real sum=0.0;
  for (int i=0; i<x.length(); ++i)
    sum += abs2(x(i)-y(i));
  return sum;
}
Real L2Norm(const ComplexColumnVector& x) {
  return sqrt(L2Norm2(x));
}

Real L2Dist(const ComplexColumnVector& x,  const ComplexColumnVector& y) {
  return sqrt(L2Dist2(x,y));
}

void operator*=(ComplexColumnVector& x, Complex c) {
  for (int i=0; i<x.length(); ++i)
    x(i) *= c;
}


void fixDiri(ChebyCoeff& f) {
  Real fa = f.eval_a();
  Real fb = f.eval_b();
  Real mean = 0.5*(fb+fa);
  Real slop = 0.5*(fb-fa);
  f[0] -= mean;
  f[1] -= slop;
}

void fixDiriMean(ChebyCoeff& f) {
  Real fa = f.eval_a();
  Real fb = f.eval_b();
  Real fm = f.mean();
  f[0] -= 0.125*(fa+fb) + 0.75*fm;
  f[1] -= 0.5*(fb-fa);
  f[2] -= 0.375*(fa+fb) - 0.75*fm;
}

void fixDiriNeum(ChebyCoeff& f)  {

  Real ya = f.a();
  Real yb = f.b();
  f.setBounds(-1,1);
  Real a = f.eval_a();
  Real b = f.eval_b();
  Real c = f.slope_a();
  Real d = f.slope_b();

  // The numercial coeffs are inverse of the matrix (values found with Maple)
  // T0(-1)  T1(-1)  T2(-1)  T3(-1)     s0      a
  // T0(1)   T1(1)   T2(1)   T3(1)      s1      b
  // T0'(-1) T1'(-1) T2'(-1) T3'(-1)    s2  ==  c
  // T0'(1)  T1'(1)  T2'(1)  T3'(1)     s3      d

  Real s0 = 0.5*(a + b) + 0.125*(c - d);
  Real s1 = 0.5625*(b - a) - 0.0625*(c + d);
  Real s2 = 0.125*(d - c);
  Real s3 = 0.0625*(a - b + c + d);

  f[0] -= s0;
  f[1] -= s1;
  f[2] -= s2;
  f[3] -= s3;
  f.setBounds(ya,yb);
}

void fixDiri(ComplexChebyCoeff& f) {
  fixDiri(f.re);
  fixDiri(f.im);
}
void fixDiriMean(ComplexChebyCoeff& f) {
  fixDiriMean(f.re);
  fixDiriMean(f.im);
}

void fixDiriNeum(ComplexChebyCoeff& f) {
  fixDiriNeum(f.re);
  fixDiriNeum(f.im);
}


void field2vector(const FlowField& u, ColumnVector& a) {

  assert(u.xzstate() == Spectral && u.ystate() == Spectral);
  int Kx = u.kxmaxDealiased();
  int Kz = u.kzmaxDealiased();
  int Ny = u.Ny();

  int N = 4*(Kx+2*Kx*Kz+Kz)*(Ny-3)+2*(Ny-2);
  if (a.length() == 0)
    a.resize_and_fill(N,0);
  else
    setToZero(a);
  assert(a.length() >= N);


  int n=0;

  // These loops follow Gibson, Halcrow, Cvitanovic table 1.
  for (int ny=2; ny<Ny; ++ny)
    a(n++) = Re(u.cmplx(0,ny,0,0)); // Ny-2 modes (line 1 of table)

  for (int ny=2; ny<Ny; ++ny)
    a(n++) = Re(u.cmplx(0,ny,0,2)); // Ny-2 modes (line 2)

  // Some coefficients from the FlowField are linearly dependent due
  // to BCs and divergence constraint. Omit these from a(n), and reconstruct
  // them in vector2field(v,u) using the constraint equations. In what follows,
  // which coefficients are omitted is determined by the changing ny loop index
  // e.g. (ny=3; n<Ny-1; ++ny) omits ny=0,1 and Ny-1, and the constraint eqns
  // that allow their reconstruction are mentioned in comments to the right
  // of the loop. 2007-06-07 jfg.
  for (int kx=1; kx<=Kx; ++kx) {
    int mx = u.mx(kx);
    for (int ny=2; ny<Ny; ++ny) {      // w BCs => ny=0,1 coefficients
      a(n++) = Re(u.cmplx(mx,ny,0,2)); // J(Ny-2) modes (line 3a)
      a(n++) = Im(u.cmplx(mx,ny,0,2)); // J(Ny-2) modes (line 3b)
    }
    for (int ny=3; ny<Ny-1; ++ny) {    // u BCs => 0,1; v BC => 2; div => Ny-1
      a(n++) = Re(u.cmplx(mx,ny,0,0)); // J(Ny-4) modes (line 4a)
      a(n++) = Im(u.cmplx(mx,ny,0,0)); // J(Ny-4) modes (line 4b)
    }
  }
  for (int kz=1; kz<=Kz; ++kz) {
    int mz = u.mz(kz);
    for (int ny=2; ny<Ny; ++ny) {      // u BCs => 0,1; v BC => 2; div => Ny-1
      a(n++) = Re(u.cmplx(0,ny,mz,0)); // K(Ny-2)  (line 5a)
      a(n++) = Im(u.cmplx(0,ny,mz,0)); // K(Ny-2)  (line 5b)
    }
    for (int ny=3; ny<Ny-1; ++ny) {    // w BCs => 0,1; v BC => 2; div => Ny-1
      a(n++) = Re(u.cmplx(0,ny,mz,2)); // K(Ny-2)  (line 6a)
      a(n++) = Im(u.cmplx(0,ny,mz,2)); // K(Ny-2)  (line 6b)
    }
  }
  for (int kx=-Kx; kx<=Kx; ++kx) {
    if (kx==0)
      continue;

    int mx = u.mx(kx);
    for (int kz=1; kz<=Kz; ++kz) {
      int mz = u.mz(kz);
      for (int ny=2; ny<Ny; ++ny) {     // u BCs => 0,1;
	a(n++) = Re(u.cmplx(mx,ny,mz,0)); // JK(Ny-2)  (line 7a)
	a(n++) = Im(u.cmplx(mx,ny,mz,0)); // JK(Ny-2)  (line 7b)
      }
      for (int ny=3; ny<Ny-1; ++ny) { // w BCs => 0,1; v BC => 2; div => Ny-1
	a(n++) = Re(u.cmplx(mx,ny,mz,2)); // JK(Ny-4)  (line 8a)
	a(n++) = Im(u.cmplx(mx,ny,mz,2)); // JK(Ny-4)  (line 8b)
      }
    }
  }
}

void vector2field(const ColumnVector& a, FlowField& u) {

  assert(u.xzstate() == Spectral && u.ystate() == Spectral);
  int Kx=u.kxmaxDealiased();
  int Kz=u.kzmaxDealiased();
  int Ny=u.Ny(); // max polynomial order == number y gridpoints - 1
  Real ya = u.a();
  Real yb = u.b();
  Real Lx = u.Lx();
  Real Lz = u.Lz();

  assert(a.length() >= 4*(Kx+2*Kx*Kz+Kz)*(Ny-3)+2*(Ny-2));

  u.setToZero();

  Complex zero(0.0, 0.0);
  ComplexChebyCoeff f0(Ny, ya, yb, Spectral);
  ComplexChebyCoeff f1(Ny, ya, yb, Spectral);
  ComplexChebyCoeff f2(Ny, ya, yb, Spectral);

  // These loops follow Gibson, Halcrow, Cvitanovic table 1.
  // The vector contains the linearly independent real numbers in a
  // zero-div no-slip FlowField. The idea is to extract these from the
  // vector and reconstruct the others from the BCs and div constraint
  // See corresponding comments in field2vector(u,v);

  int n=0;

  // =========================================================
  // (0,0) Fourier mode
  for (int ny=2; ny<Ny; ++ny)
    f0.re[ny] = a(n++);
  fixDiri(f0.re);
  for (int ny=0; ny<Ny; ++ny)
    u.cmplx(0,ny,0,0) = f0[ny];

  for (int ny=2; ny<Ny; ++ny)
    f2.re[ny] = a(n++);
  fixDiri(f2.re);
  for (int ny=0; ny<Ny; ++ny)
    u.cmplx(0,ny,0,2) = f2[ny];

  // =========================================================
  // (kx,0) Fourier modes, 0<kx
  for (int kx=1; kx<=Kx; ++kx) {
    int mx = u.mx(kx);

    for (int ny=2; ny<Ny; ++ny) {
      f2.re[ny] = a(n++);
      f2.im[ny] = a(n++);
    }
    fixDiri(f2);

    for (int ny=3; ny<Ny-1; ++ny) {
      f0.re[ny] = a(n++);
      f0.im[ny] = a(n++);
    }
    f0.re[Ny-1] = 0.0;
    f0.im[Ny-1] = 0.0;
    fixDiriMean(f0);

    integrate(f0, f1);
    f1.sub(0, 0.5*(f1.eval_a() + f1.eval_b()));
    f1 *= Complex(0.0, -(2*pi*kx)/Lx);

    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mx,ny,0,0) = f0[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mx,ny,0,1) = f1[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mx,ny,0,2) = f2[ny];

    // ------------------------------------------------------
    // Now copy conjugates of u(kx,ny,0,i) to u(-kx,ny,0,i). These are
    // redundant modes stored only for the convenience of FFTW.
    int mxm = u.mx(-kx);
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mxm,ny,0,0) = conj(f0[ny]);
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mxm,ny,0,1) = conj(f1[ny]);
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(mxm,ny,0,2) = conj(f2[ny]);

  }


  // =========================================================
  // (0,kz) Fourier modes, 0<kz.
  for (int kz=1; kz<=Kz; ++kz) {
    int mz = u.mz(kz);

    for (int ny=2; ny<Ny; ++ny) {
      f0.re[ny] = a(n++);
      f0.im[ny] = a(n++);
    }
    fixDiri(f0);
    for (int ny=3; ny<Ny-1; ++ny) {
      f2.re[ny] = a(n++);
      f2.im[ny] = a(n++);
    }
    f2.re[Ny-1] = 0.0;
    f2.im[Ny-1] = 0.0;
    fixDiriMean(f2);

    integrate(f2, f1);
    f1.sub(0, 0.5*(f1.eval_a() + f1.eval_b()));
    f1 *= Complex(0.0, -(2*pi*kz)/Lz);

    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(0,ny,mz,0) = f0[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(0,ny,mz,1) = f1[ny];
    for (int ny=0; ny<Ny; ++ny)
      u.cmplx(0,ny,mz,2) = f2[ny];
  }

  // =========================================================
  for (int kx=-Kx; kx<=Kx; ++kx) {
    if (kx==0)
      continue;

    int mx = u.mx(kx);
    for (int kz=1; kz<=Kz; ++kz) {
      int mz = u.mz(kz);

      for (int ny=2; ny<Ny; ++ny) {
	f0.re[ny] = a(n++);
	f0.im[ny] = a(n++);
      }
      fixDiri(f0);

      // f0 is complete. Load it into u.
      for (int ny=0; ny<Ny; ++ny)
	u.cmplx(mx,ny,mz,0) = f0[ny];

      for (int ny=3; ny<Ny-1; ++ny) {
	f2.re[ny] = a(n++);
	f2.im[ny] = a(n++);
      }
      f2.re[Ny-1] = -f0.re[Ny-1]*(kx*Lz)/(kz*Lx);
      f2.im[Ny-1] = -f0.im[Ny-1]*(kx*Lz)/(kz*Lx);

      // Adjust 0,1,2 coeffs of f2 so that f2(+/-1) = 0 and
      // kz/Lz mean(f2) + kx/Lx mean(f0) == 0 (so that  v(1)==v(-1))
      Complex f2a = f2.eval_a();
      Complex f2b = f2.eval_b();
      Complex f0m = f0.mean();
      Complex f2m = f2.mean() + (kx*Lz)/(kz*Lx)*f0m;

      f2.sub(0, 0.125*(f2a+f2b) + 0.75*f2m);
      f2.sub(1, 0.5*(f2b-f2a));
      f2.sub(2, 0.375*(f2a+f2b) - 0.75*f2m);

      for (int ny=0; ny<Ny; ++ny)
	u.cmplx(mx,ny,mz,2) = f2[ny];

      f0 *= Complex(0, -2*pi*kx/Lx);
      f2 *= Complex(0, -2*pi*kz/Lz);
      f0 += f2;

      integrate(f0, f1);
      f1.sub(0, 0.5*(f1.eval_a() + f1.eval_b()));

      for (int ny=0; ny<Ny; ++ny)
	u.cmplx(mx,ny,mz,1) = f1[ny];
    }
  }
}

void fixdivnoslip(FlowField& u) {
  ColumnVector v;
  field2vector(u,v);
  vector2field(v,u);
}

void field2vector(const FlowField& u, ColumnVector& v, const std::vector<RealProfileNG>& basis) {
  if (basis.size() == 0)
    field2vector(u,v);
  else {
    if (v.length() == 0)
      v.resize_and_fill(basis.size(), 0);
    else
      setToZero(v);

    for (uint i=0; i<basis.size(); ++i)
      v(i) = L2InnerProduct(u, basis[i]);
  }
  return;
}


void vector2field(const ColumnVector& v, FlowField& u, const std::vector<RealProfileNG>& basis) {
  if (basis.size() == 0)
    vector2field(v,u);
  else {

    RealProfileNG prof;
    const Real EPSILON = 1e-16;
    u.setToZero();

    for (uint i=0; i<basis.size(); ++i) {
      if (abs(v(i)) < EPSILON)
	continue;
      prof = basis[i];
      prof *= v(i);
      u += prof;
    }
  }
}

void vector2field(const vector<RealProfile>& basis, const ColumnVector& a,
		  FlowField& u) {
  int N = basis.size();
  assert(a.length() >= N);
  u.setToZero();
  RealProfile tmp;

  const Real EPSILON = 1e-16;
  for (int n=0; n<N; ++n) {
    Real an = a(n);
    if (abs(an) < EPSILON)
      continue;

    tmp = basis[n];
    tmp *= an;

    assert(tmp.kx() <= u.kxmax() && tmp.kx() >= u.kxmin() &&
	   tmp.kz() <= u.kzmax() && tmp.kz() >= u.kzmin());

    u += tmp;
  }
}




void field2vector(const vector<RealProfile>& basis, const FlowField& u, ColumnVector& a) {

  int N = basis.size();
  if (a.length() == 0)
    a.resize_no_fill(N);
  assert(a.length() >= N);

  for (int n=0; n<N; ++n)
    a(n) = L2InnerProduct(u, basis[n]);
}



void vector2field(const vector<BasisFunc>& basis, const ComplexColumnVector& a,
		  FlowField& u) {
  int N = basis.size();
  assert(a.length() >= N);
  u.setToZero();
  BasisFunc tmp;

  const Real EPSILON = 1e-16;
  for (int n=0; n<N; ++n) {
    Complex an = a(n);
    if (abs(an) < EPSILON)
      continue;

    tmp = basis[n];
    tmp *= an;

    assert(tmp.kx() <= u.kxmax() && tmp.kx() >= u.kxmin() &&
	   tmp.kz() <= u.kzmax() && tmp.kz() >= u.kzmin());

    u += tmp;
  }
}

void quicksort(ComplexColumnVector& z) {
  int N = z.length();
  array<Real> x(N);
  for (int n=0; n<N; ++n)
    x[n] = abs(z(n));
  array<int> i = quicksort(x);
  ComplexColumnVector w(N);
  for (int n=0; n<N; ++n)
    w(n) = z(i[N-n-1]);
  z = w;
}

// Does extra copying
void quicksort(ComplexColumnVector& Lambda, ComplexMatrix& V) {
  int M = V.rows();
  int N = V.cols();
  assert(Lambda.length() == N);

  array<Real> magn(N);
  for (int n=0; n<N; ++n)
    magn[n] = abs(Lambda(n));
  array<int> i = quicksort(magn);

  ComplexColumnVector Lambda_sorted(N);
  ComplexMatrix V_sorted(M,N);

  for (int n=0; n<N; ++n) {
    Lambda_sorted(n) = Lambda(i[N-n-1]);
    V_sorted.insert(V.column(i[N-n-1]),0,n);
  }
  Lambda = Lambda_sorted;
  V = V_sorted;
}

void leastsquares(const Matrix& Hn, ColumnVector& bk, ColumnVector& y,
		  Real& residual) {
  //cout << "entering leastsquares..." << endl;

  // Hn y = bk, Hn = QR
  // QQR y = bk
  //  R y = Q' bk;

  QR qr(Hn, QR::economy);
  Matrix Q = qr.Q();
  Matrix R = qr.R();
  int N = R.rows();
  RowVector c = bk.transpose()*Q;

  // Backsolve R y = Q* b
  for (int i=N-1; i>=0; --i) {
    Real sum = c(i);
    for (int j=i+1; j<N; ++j)
      sum -= R(i,j)*y(j);
    y(i) = sum/R(i,i);
  }

  residual = L2Norm(Hn*y-bk)/L2Norm(bk);
  //cout << "subspace residual = " << L2Norm(R*y-c)/L2Norm(c) << endl;
  //cout << "leastsqs residual = " << residual << endl;
  //cout << "L2Norm(bk)        == " << L2Norm(bk) << endl;
  //cout << "L2Norm(Hn*y-bk)   == " << L2Norm(Hn*y-bk) << endl;
  //cout << "L2Norm(Q'*bk)     == " << L2Norm(c) << endl;
  //cout << "L2Norm(R*y-Q'*bk) == " << L2Norm(R*y-c) << endl;
}

Arnoldi::Arnoldi()
  :
  M_(0),
  Niter_(0),
  n_(0),
  condition_(0),
  H_(0,0),
  Q_(0,0),
  Vn_(0,0),
  qn_(0),
  v_(0),
  ew_(0),
  ev_(0,0)
{}

Arnoldi::Arnoldi(const ColumnVector& b, int Niterations, Real minCondition)
  :
  M_(b.length()),
  Niter_(Niterations),
  n_(0),
  condition_(minCondition),
  H_(Niter_+1, Niter_),
  Q_(M_, Niter_+1),
  Vn_(0,0),
  qn_(b),
  v_(M_),
  ew_(0),
  ev_(0,0)
{
  setToZero(H_);
  setToZero(Q_);
  qn_ *= 1.0/L2Norm(qn_);
  qn_ *= 1.0/L2Norm(qn_);
  Q_.insert(qn_,0,0);
}

void Arnoldi::iterate(const ColumnVector& Aq) {
  if (n_ == Niter_) {
    cerr << "warning : Arnoldi::iterate(Aq) : \n"
	 << "reached maximum number of iterations. doing nothing.\n";
  }
  v_ = Aq;

  // Orthogonalize v and insert it in Q matrix
  //cout << "calculating arnoldi iterate..." << endl;
  for (int j=0; j<=n_; ++j) {
    //cout << j << ' ' << flush;
    ColumnVector Qj = Q_.column(j);
    H_(j,n_) = Qj.transpose() * v_;
    v_ -= H_(j,n_)*Qj;
  }
  //cout << endl;
  Real vnorm = L2Norm(v_);

  if (abs(vnorm) < condition_) {
    H_(n_+1, n_) = 0.0;
    v_ *= 0.0;
    cerr << "Arnoldi breakdown. Exiting\n";
    exit(1);
  }

  else {
    H_(n_+1, n_) = vnorm;
    v_ *= 1.0/vnorm;
  }
  Q_.insert(v_,0,n_+1);
  qn_ = Q_.column(n_+1);
  ++n_;
}

const ColumnVector& Arnoldi::testVector() const {
  return qn_;
}

int Arnoldi::n() const {return n_;}
int Arnoldi::Niter() const {return Niter_;}

const ComplexColumnVector& Arnoldi::ew() {
  if (ew_.length() != n_)
    eigencalc();
  return ew_;
}

void Arnoldi::eigencalc() {
  Matrix Qn  = Q_.extract(0,0,M_-1,n_-1);
  Matrix Hn = H_.extract(0,0,n_-1,n_-1);
  //cout << "Q == \n" << Q_ << endl;
  //cout << "H == \n" << H_ << endl;
  //cout << "Hn == \n" << Hn << endl;
  //  eignHn_ = EIG(Hn);
  EIG eignHn(Hn);
  ew_ = eignHn.eigenvalues();
  ComplexMatrix W = eignHn.eigenvectors();
  quicksort(ew_, W);
  ev_ = Qn*W;
};

const ComplexMatrix& Arnoldi::ev() {
  if (ev_.length() != n_)
    eigencalc();
  return ev_;
}


GMRES::GMRES()
  :
  M_(0),
  Niter_(0),
  n_(0),
  condition_(0),
  H_(0,0),
  Q_(0,0),
  Vn_(0,0),
  x0_(0),
  v_(0),
  qn_(0),
  bnorm_(0)
{}

GMRES::GMRES(const ColumnVector& b, int Niterations, Real minCondition)
  :
  M_(b.length()),
  Niter_(Niterations),
  n_(0),
  condition_(minCondition),
  H_(Niter_+1, Niter_),
  Q_(M_, Niter_+1),
  Vn_(0,0),
  x0_(0),
  v_(0),
  qn_(b),
  bnorm_(L2Norm(b))
{
  setToZero(H_);
  setToZero(Q_);
  qn_ *= 1.0/L2Norm(qn_);
  qn_ *= 1.0/L2Norm(qn_);
  Q_.insert(qn_,0,0);
}

/**********************************
GMRES::GMRES(const ColumnVector& guess, const ColumnVector& b, int Niterations, Real minCondition)
  :
  M_(b.length()),
  Niter_(Niterations),
  n_(0),
  condition_(minCondition),
  H_(Niter_+1, Niter_),
  Q_(M_, Niter_+1),
  Vn_(0,0),
  x0_(guess),
  v_(0),
  qn_(b),
  bnorm_(L2Norm(b))
{
  if (guess.length() != b.length()) {
    cerr << "error in : GMRES::GMRES(guess,b,Niterations,minCondition)" << endl;
    cerr << "guess.length() != b.length()). Exiting." << endl;
  }
  setToZero(H_);
  setToZero(Q_);
  qn_ *= 1.0/L2Norm(qn_);
  qn_ *= 1.0/L2Norm(qn_);
  Q_.insert(qn_,0,0);
}
************************************/

void GMRES::iterate(const ColumnVector& Aq) {
  if (n_ == Niter_) {
    cerr << "warning : GMRES::iterate(Ab) : \n"
	 << "reached maximum number of iterations. doing nothing\n";
    return;
  }
  v_ = Aq;

  // Orthogonalize v and insert it in Q matrix
  //cout << "calculating arnoldi iterate..." << endl;
  for (int j=0; j<=n_; ++j) {
    //cout << j << ' ' << flush;
    ColumnVector Qj = Q_.column(j);
    H_(j,n_) = Qj.transpose() * v_;
    v_ -= H_(j,n_)*Qj;
  }
  //cout << endl;
  Real vnorm = L2Norm(v_);

  // If v is within subspace spanned by Q, replace it with a random vector.
  // There's a measure 0 chance that any particular random vector will also
  // spanned by Q, so allow for a limited number of retries.
  int maxretries = 10;
  int retries = 0;
  while (abs(vnorm) < condition_ && retries++ < maxretries) {
    cerr << "GMRES benign breakdown: \n"
	 << "  Using random vector instead of linearly dependent Krylov vector\n.";

    for (int i=0; i<v_.length(); ++i)
      v_(i) = randomReal();

    // Gram-Schmidt orthogonalize v w.r.t Q
    for (int j=0; j<=n_; ++j) {
      ColumnVector Qj = Q_.column(j);
      H_(j,n_) = Qj.transpose() * v_;
      v_ -= H_(j,n_)*Qj;
    }
    //cout << endl;
    vnorm = L2Norm(v_);
  }

  H_(n_+1, n_) = vnorm;
  v_ *= 1.0/vnorm;
  Q_.insert(v_,0,n_+1);
  qn_ = Q_.column(n_+1);

  Matrix Hn = H_.extract(0,0,n_+1,n_);
  Matrix Qn = Q_.extract(0,0,M_-1,n_);
  ColumnVector bk(n_+2);
  setToZero(bk);
  bk(0) = bnorm_;

  ColumnVector y(n_+1);
  leastsquares(Hn,bk,y,residual_);
  xn_ = Qn*y;
  ++n_;
}

int GMRES::n() const {return n_;}
int GMRES::Niter() const {return Niter_;}

const ColumnVector& GMRES::testVector() const {
  return qn_;
}
const ColumnVector& GMRES::solution() const{
  return xn_;
}
Real GMRES::residual() const{
  return residual_;
}
Matrix GMRES::Hn() const {
  return H_.extract(0,0,n_,n_-1);
}
Matrix GMRES::Qn() const {
  return Q_.extract(0,0,M_-1,n_-1);
}
Matrix GMRES::Qn1() const {
  return Q_.extract(0,0,M_-1,n_);
}

/******************************************
HookStepSearch::HookStepSearch()
  :
  newtonStep_(0),
  stuck_(false),
  residual_(0),
  delta_(0),
  x_(0),
  fx_(0),
{}

HookStepSearch::HookStepSearch(Rn2Rnfunc, const ColumnVector& x0);
  :
  newtonStep_(0),
  residual_(0),
  delta_(0),
  stuck_(false),
  Nunk_(0),
  x_(0),
  fx_(0),
{}


HookStepSearch::HookStepSearch(Rn2Rnfunc, const ColumnVector& x0);const HookstepParams& p) : params_(p) {
  newtonStep_(0),
  stuck_(false),
  residual_(0),
  Nunk_(0),
  delta_(0),
  x_(0),
  fx_(0),
}
****************************************************/

ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0) {
  HookstepParams p;
  int foo;
  return hookstepSearch(f, x0, p, foo);
}

HookstepParams::HookstepParams(Real epsSearch_,
			       Real epsDx_,
			       bool centdiff_,
			       int  Nnewton_,
			       int  Nhook_,
			       Real deltaInit_,
			       Real deltaMin_,
			       Real deltaMax_,
			       Real deltaFuzz_,
			       Real lambdaMin_,
			       Real lambdaMax_,
			       Real lambdaRequiredReduction_,
			       Real improvReq_,
			       Real improvOk_,
			       Real improvGood_,
			       Real improvAcc_,
			       bool pausing_,
			       ostream* logstream_)
  :
  epsSearch(epsSearch_),
  epsDx(epsDx_),
  centdiff(centdiff_),
  Nnewton(Nnewton_),
  Nhook(Nhook_),
  deltaInit(deltaInit_),
  deltaMin(deltaMin_),
  deltaMax(deltaMax_),
  deltaFuzz(deltaFuzz_),
  lambdaMin(lambdaMin_),
  lambdaMax(lambdaMax_),
  lambdaRequiredReduction(lambdaRequiredReduction_),
  improvReq(improvReq_),
  improvOk(improvOk_),
  improvGood(improvGood_),
  improvAcc(improvAcc_),
  xscale(0),
  fscale(0),
  pausing(pausing_),
  logstream(logstream_)
{}

ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0, const HookstepParams& p, int& nsteps) {

  Real epsSearch = p.epsSearch; // stop search if |f(x)| < epsSearch
  Real epsDx     = p.epsDx;     // scale for perturbs in finite-diff estimate of Df
  //bool centdiff  = p.centdiff;// used centered differences in estimating Df
  int  Nnewton   = p.Nnewton;   // max number of Newton step
  int  Nhook     = p.Nhook;     // max number of hookstep iterations per Newton
  Real deltaInit = p.deltaInit; // initial trust region radius
  Real deltaMin  = p.deltaMin;  // stop if radius of trust region gets this small
  Real deltaMax  = p.deltaMax;  // don't let trust region radius get bigger than this
  Real deltaFuzz = p.deltaFuzz; // accept steps within (1+/-deltaFuzz)*delta
  Real lambdaMin = p.lambdaMin; // minimum delta shrink rate
  Real lambdaMax = p.lambdaMax; // maximum delta expansion rate
  Real lambdaRequiredReduction = p.lambdaRequiredReduction;
  Real improvReq = p.improvReq;
  Real improvOk   = p.improvOk;
  Real improvGood = p.improvGood;
  Real improvAcc  = p.improvAcc;
  ColumnVector xscale = p.xscale;
  ColumnVector fscale = p.fscale;
  bool pausing = p.pausing;
  ostream* os = p.logstream;

  string outdir = "./";

  Real delta = deltaInit;
  int fcount_newton = 0;
  int fcount_hookstep = 0;

  const int Nunk = x0.length();
  const int Neqn = Nunk;
  int label;
  bool cacheable;

  ofstream xnewtos("xnewt.asc");
  ofstream dnewtos("dxnewt.asc");
  ofstream fnewtos("fnewt.asc");
  ofstream osconv("convergence.asc");
  osconv.setf(ios::left);
  osconv << setw(14) << "% residual" << setw(14) << "delta"
	 << setw(14) << "L2Norm(dx)" << setw(14) << "L2Norm(sN)"
	 << setw(14) << "L2Norm(sH)" << setw(10) << "ftotal"
	 << setw(10) << "fnewt" << setw(10) << "fhook" << endl;

  // ==============================================================
  // Initialize Newton iteration
  if (os) *os << "Computing f(x)" << endl;
  ColumnVector       x(x0);
  ColumnVector      fx = (*f)(x, label=0, cacheable=true); // 0 ==> current best guess
  ++fcount_newton;
  // From here on, scale/unscale x when interfacing with the external world
  rescale(x,xscale);

  ColumnVector      dx(Nunk); // temporary for increments in unknowns
  ColumnVector   Df_dx(Nunk); // "Df*dx" = 1/eps (f(x + eps dx)
  ColumnVector      xS(Nunk); // stores xN and xH, the Newton and Hookstep unknowns
  ColumnVector      fS(Nunk); // stores fN and fH, fS = f(xS);
  ColumnVector      sN(Nunk); // the Newton step

  setToZero(dx);
  //for (i=0; i<Nunk; ++i)
  //dx(i) = 0.0;

  Real rx = 0.0;        // Dennis & Schnabel residual r(x) = 1/2 ||f(x)||^2
  Real init_rx = 0.0;
  Real prev_rx = 0.0;
  Real snorm   = 0.0;   // norm of hookstep
  Real sNnorm  = 0.0;   // norm of newton step
  Real dxnorm  = 0.0;
  Real xnorm   = 0.0;
  bool stuck = false;

  for (int newtonStep=0; newtonStep<Nnewton; ++newtonStep) {

    nsteps = newtonStep;

    if (os) *os << "========================================================" << endl;
    if (os) *os << "Newton iteration number " << newtonStep << endl;

    // Compute quantities that will be used multiple times
    // fx     = f(x) - x
    dxnorm = L2Norm(dx);
    xnorm = L2Norm(x);
    rx = 0.5*L2Norm2(fx);

    if (os) {
      *os << "Current state of Newton iteration:" << endl;
      *os << "fcount_newton  == " << fcount_newton << endl;
      *os << "fcount_hookstep== " << fcount_hookstep << endl;
      *os << "    L2Norm(x)  == " << xnorm << endl;
      *os << "   L2Norm(dx)  == " << dxnorm <<endl;
      *os << "   L2Norm(sH)  == " << snorm <<endl;
      //os << "L2Dist(x,x0)   == " << L2Dist(x, x0) << endl;
      *os << "rx == 1/2 Norm2(f(x)) : " << endl;
      *os << "   initial  rx == " << init_rx << endl;
      *os << "   previous rx == " << prev_rx << endl;
      *os << "   current  rx == " << rx << endl;
      *os << "             x == ";
      for (int i=0; i<Nunk; ++i) *os << x(i) << ' ';  *os << endl;
      *os << "            dx == ";
      for (int i=0; i<Nunk; ++i) *os << dx(i) << ' '; *os << endl;
      *os << "            fx == ";
      for (int i=0; i<Nunk; ++i) *os << fx(i) << ' '; *os << endl;
      *os << "        delta  == " << delta << endl;
    }
    if (newtonStep == 0) {
      init_rx = rx;
      prev_rx = rx;
    }

    osconv.setf(ios::left);
    osconv.setf(ios::left); osconv << setw(14) << rx;
    osconv.setf(ios::left); osconv << setw(14) << delta;
    osconv.setf(ios::left); osconv << setw(14) << dxnorm;
    osconv.setf(ios::left); osconv << setw(14) << sNnorm;
    osconv.setf(ios::left); osconv << setw(14) << snorm;
    osconv.setf(ios::left); osconv << setw(10) << fcount_newton + fcount_hookstep;
    osconv.setf(ios::left); osconv << setw(10) << fcount_newton;
    osconv.setf(ios::left); osconv << setw(10) << fcount_hookstep;
    osconv << endl;

    unscale( x,xscale);
    unscale(dx,xscale);
    save(x, outdir + "xnewt" + i2s(newtonStep));
    save(x, outdir + "xbest");
    for (int i=0; i<Nunk; ++i) {
      xnewtos << x(i) << ' ';
      dnewtos << dx(i) << ' ';
      fnewtos << fx(i) << ' ';
    }
    xnewtos << endl;
    dnewtos << endl;
    fnewtos << endl;
    rescale( x,xscale);
    rescale(dx,xscale);

    if (rx < epsSearch) {
      if (os) *os << "Newton search converged. Breaking." << endl;
      if (os) *os << "1/2 L2Norm(f(x)) == " << rx << " < "
	   << epsSearch << " == " << epsSearch << endl;
      break;
    }
    else if (stuck) {
      if (os) *os << "Newton search is stuck. Breaking." << endl;
      break;
    }
    else if (newtonStep == Nnewton) {
      if (os) *os << "Reached maximum number of Newton steps. Breaking." << endl;
      break;
    }
    if (pausing)
      cfpause();

    prev_rx = rx;

    // Set up RHS vector b = -f(x)
    ColumnVector b(Neqn);
    setToZero(b);
    b = fx; // field2vector(fu, b, basis);
    b *= -1.0;

    // Estimate f_ij = df_i/dx_j with finite differences
    //               = 1/eps (f_i(x + eps e_j) - f_i(x))
    Matrix Df(Nunk, Nunk);

    for (int j=0; j<Nunk; ++j) {
      ColumnVector f_x_dxj(Nunk);
      ColumnVector x_dxj = x;
      x_dxj(j) += epsDx;
      unscale(x_dxj, xscale);
      f_x_dxj = (*f)(x_dxj, label=-1, cacheable=false); // -1 ==> evaluating Df

      for (int i=0; i<Nunk; ++i)
	Df(i,j) = (f_x_dxj(i)- fx(i))/epsDx;
    }
    fcount_newton += Nunk*Nunk;

    save(Df, "Df");

    SVD svd(Df, SVD::std);

    Matrix V = svd.right_singular_matrix();
    Matrix Ut = svd.left_singular_matrix().transpose();
    DiagMatrix D = svd.singular_values();
    if (os) {
      *os << "D == ";
      for (int i=0; i<D.rows(); ++i)
	*os << D(i,i) << ' ';
      *os << endl;
    }
    //cfpause();

    solve(Ut, D, V, b, sN);
    sNnorm = L2Norm(sN);

    if (os) *os << "newton step sN == " << sN(0) << ' ' << sN(1) << endl;



    // ==================================================================
    // Hookstep algorithm

    if (os) *os << "------------------------------------------------" << endl;
    if (os) *os << "Beginning hookstep calculations." << endl;

    int Nk = Nunk;

    ColumnVector bh = Ut*b; // b hat

    int hookcount = 0;
    bool have_backup  = false;
    Real backup_snorm = 0.0;
    Real backup_delta = 0.0;
    Real backup_rS    = 0.0;

    ColumnVector backup_dx;
    ColumnVector backup_fS;

    ColumnVector sH;

    //bool deltahas_decreased = false;
    Real deltaMaxLocal = deltaMax;

    // Find a good trust region and the optimal hookstep in it.
    // Quality of trust region
    while (true) {

      if (os) *os << "-------------------------------------------" << endl;
      if (os) *os << "Newton, hookstep number " << newtonStep << ", "
	   << hookcount++ << endl;
      if (os) *os << "delta == " << delta << endl;

      bool hookstep_equals_newtonstep;

      if (L2Norm(sN) <= delta) {
	if (os) *os << "Newton step is within trust region: " << endl;
	if (os) *os << "L2Norm(sN) == " << L2Norm(sN) << " <= " << delta << " == delta" << endl;
	hookstep_equals_newtonstep = true;
	sH = sN;
      }
      else {
	if (os) *os << "Newton step is outside trust region: " << endl;
	if (os) *os << "L2Norm(sN) == " << L2Norm(sN) << " > " << delta << " == delta" << endl;
	if (os) *os << "Calculate hookstep sH with radius L2Norm(sH) == delta" << endl;

	hookstep_equals_newtonstep = false;

	// This for-loop determines the hookstep. Search for value of mu that
	// produces ||shmu|| == delta. That provides optimal reduction of
	// quadratic model of residual in trust region norm(shmu) <= delta.
	Real mu = 0.0;
	Real nshmu = 0;
	ColumnVector shmu(Nk); // (s hat)(mu)]

	for (int hookSearch=0; hookSearch<Nhook; ++hookSearch) {
	  for (int i=0; i<Nk; ++i)
	    shmu(i) = bh(i)/(D(i,i) + mu); // research notes
	  nshmu = L2Norm(shmu);
	  Real Phi = nshmu*nshmu - delta*delta;
	  if (os) *os << "mu, L2Norm(sH(mu)) == " << mu << ", " << nshmu << endl;

	  // Found satisfactory value of mu and thus shmu and sH s.t. |sH| < delta
	  if (nshmu < delta || (nshmu > (1-deltaFuzz)*delta && nshmu < (1+deltaFuzz)*delta))
	    break;

	  // Update value of mu and try again. Update rule is a Newton search for
	  // Phi(mu)==0 based on a model of form a/(b+mu) for norm(sh(mu)). See
	  // Dennis & Schnabel.
	  else if (hookSearch<Nhook-1) {
	    Real PhiPrime = 0.0;
	    for (int i=0; i<Nk; ++i) {
	      Real di_mu = D(i,i) + mu;
	      Real bi = bh(i);
	      PhiPrime -= 2*bi*bi/(di_mu*di_mu*di_mu);
	    }
	    mu -= (nshmu/delta) * (Phi/PhiPrime);
	  }
	  else {
	    if (os) *os << "Couldn't find solution of hookstep optimization eqn Phi(mu)==0" << endl;
	    if (os) *os << "This shouldn't happen. It indicates an error in the algorithm." << endl;
	    if (os) *os << "Exiting.\n";
	    exit(1);
	  }
	} // search over mu for norm(s) == delta

	if (os) *os << "Found hookstep of proper radius" << endl;
	if (os) *os << "nshmu == " << L2Norm(shmu) << " ~= " << delta << " == delta " << endl;
	sH = V*shmu;
      } // else {Newton step outside trust region, so

      // Compute
      // (1) actual residual from evaluation of
      //     rS == r(s+ds) act  == 1/2 (f(x+dx),f(x+dx))
      //
      // (2) predicted residual from quadratic model of r(s)
      //     rP == r(s+ds) pred  == 1/2 (f(x),f(x)) + (f(x), Df dx)
      //
      // (3) slope of r(s)
      //     dr/ds == (r(s + ds) - r(s))/|ds|
      //           == (f(x,T), Df dx /|ds|
      // where ( , ) is Vector inner product V2IP, which equals L2 norm of vector rep

      // (1) Compute actual residual of step, rS
      if (os) *os << "Computing residual of hookstep sH" << endl;
      dx = sH;
      dxnorm = L2Norm(dx);
      xS  = x;
      xS += dx;

      if (os) *os << "L2Norm(dx) == " << L2Norm(dx) << endl;
      unscale(xS, xscale);
      fS = (*f)(xS, newtonStep+1, cacheable=true);
      ++fcount_hookstep;
      rescale(xS, xscale);

      if (os) *os << endl;

      Real rS = 0.5*L2Norm2(fS);       // actual residual of hookstep
      Real Delta_rS = rS - rx;         // improvement in residual
      //if (os) *os << "r(s), r(s+ds) == " << rx << ", " << rS << endl;

      // (2) and (3) Compute quadratic model and slope
      if (os) *os << "Computing local quadratic model of residual" << endl;
      ColumnVector Df_dx = Df*dx;
      if (os) *os << endl;
      // Local quadratic and linear models of residual, based on Taylor exp at current position
      // Quadratic model of r(s+ds): rQ == 1/2 |f(x) + Df dx|^2
      // Linear  model   of r(s+ds): rL == 1/2 (f(x),f(x)) + (f(x), Df dx)

      Real Delta_rL = L2IP(fx,Df_dx); // rL  - 1/2 (f(x),f(x)) == (f(x), Df dx)
      Real rL = rx + Delta_rL;        // rL == 1/2 (f(x),f(x))  + (f(x), Df dx)

      if (rL >= rx) {
	if (os) *os << "error : local linear model of residual is increasing, indicating\n"
		    << "        that the solution to the Newton equations is inaccurate\n"
		    << "returning best answer so far" << endl;
	return x;
      }

      Df_dx += fx;
      Real rQ = 0.5*L2Norm2(Df_dx);  // rQ == 1/2 |f(x) + Df dx|^2
      Real Delta_rQ = rQ - rx;       // rQ == 1/2 |(f(x)|^2 - 1/2 |f(x) + Df dx|^2

      snorm = L2Norm(sH);

      // ===========================================================================
      if (os) *os << "Determining what to do with current Newton/hookstep and trust region" << endl;

      // Coefficients for non-local quadratic model of residual, based on r(s), r'(s) and r(d+ds)
      // dr/ds == (f(x), Df dx)/|ds|
      Real drds  = Delta_rL/snorm;

      // Try to minimize a quadratic model of the residual
      // r(s+ds) = r(s) + r' |ds| + 1/2 r'' |ds|^2
      Real lambda = -0.5*drds*snorm/(rS - rx - drds*snorm);

      // Compare the actual reduction in residual to the quadratic model of residual.
      // How well the model matches the actual reduction will determine, later on, how and
      // when the radius of the trust region should be changed.
      Real Delta_rS_req  = improvReq*drds*snorm;    // the minimum acceptable change in residual
      Real Delta_rS_ok   = improvOk*Delta_rQ;       // acceptable, but reduce trust region for next newton step
      Real Delta_rS_good = improvGood*Delta_rQ;     // acceptable, keep same trust region in next newton step
      Real Delta_rQ_acc  = abs(improvAcc*Delta_rQ); // for accurate models, increase trust region and recompute hookstep
      Real Delta_rS_accP = Delta_rQ + Delta_rQ_acc; //   upper bound for accurate predictions
      Real Delta_rS_accM = Delta_rQ - Delta_rQ_acc; //   lower bound for accurate predictions

      // Characterise change in residual Delta_rS
      ResidualImprovement improvement;  // fine-grained characterization

      // Place improvement in contiguous spectrum: Unacceptable > Poor > Ok > Good.
      if (Delta_rS > Delta_rS_req)
	improvement = Unacceptable;     // not even a tiny fraction of linear rediction
      else if (Delta_rS > Delta_rS_ok)
	improvement = Poor;             // worse than small fraction of quadratic prediction
      else if (Delta_rS < Delta_rS_ok && Delta_rS > Delta_rS_good)
	improvement = Ok;
      else {
	improvement = Good;             // not much worse or better than large fraction of prediction
	if (Delta_rS_accM <= Delta_rS &&  Delta_rS <= Delta_rS_accP)
	  improvement = Accurate;         // close to quadratic prediction
	else if (Delta_rS < Delta_rL)
	  improvement = NegaCurve;        // negative curvature in r(|s|) => try bigger step
      }
      int w = 14;

      if (os) *os << " rx      == " << setw(w) << rx            << " residual at current position" << endl;
      if (os) *os << " rS      == " << setw(w) << rS            << " residual of newton/hookstep" << endl;
      if (os) *os << "Delta_rS == " << setw(w) << Delta_rS      << " actual improvement in residual from newton/hookstep" << endl;
      if (os) *os << endl;

      if (improvement==Unacceptable)
      if (os) *os << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Unacceptable <------ " << endl;
      if (os) *os << "            " << setw(w) << Delta_rS_req  << " lower bound for acceptable improvement" << endl;
      if (improvement==Poor)
      if (os) *os << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Poor <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rS_ok   << " upper bound for ok improvement." << endl;
      if (improvement==Ok)
      if (os) *os << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Ok <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rS_good << " upper bound for good improvement." << endl;
      if (improvement==Good)
      if (os) *os << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Good <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rS_accP << " upper bound for accurate prediction." << endl;
      if (improvement==Accurate)
      if (os) *os << "Delta_rS == " << setw(w) << Delta_rS      << " ------> Accurate <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rS_accM << " lower bound for accurate prediction." << endl;
      if (os) *os << "            " << setw(w) << Delta_rL      << " local linear model of improvement" << endl;
      if (improvement==NegaCurve)
      if (os) *os << "Delta_rS == " << setw(w) << Delta_rS      << " ------> NegativeCurvature <------" << endl;

      bool recompute_hookstep = false;

      if (os) *os << "lambda       == " << lambda << " is the reduction/increase factor for delta suggested by quadratic model" << endl;
      if (os) *os << "lambda*delta == " << lambda*delta << " is the delta suggested by quadratic model" << endl;

      // See comments at end for outline of this control structure

      // Following Dennis and Schnable, if the Newton step lies within the trust region,
      // reset trust region radius to the length of the Newton step, and then adjust it
      // further according to the quality of the residual.
      //if (hookstep_equals_newtonstep)
      //delta = snorm;

      // CASE 1: UNACCEPTABLE IMPROVEMENT
      // Improvement is so bad (relative to gradient at current position) that we'll
      // in all likelihood get better results in a smaller trust region.
      if (improvement == Unacceptable) {
	if (os) *os << "Improvement is unacceptable." << endl;

	if (have_backup) {
	  if (os) *os << "But we have a backup step that was acceptable." << endl;
	  if (os) *os << "Revert to backup step and backup delta, take the step, and go to next Newton iteration." << endl;
	  dx = backup_dx;
	  fS = backup_fS;
	  rS = backup_rS;

	  snorm = backup_snorm;
	  delta = backup_delta;

	  recompute_hookstep = false;
	}
	else {
	  if (os) *os << "No backup step is available." << endl;
	  if (os) *os << "Reduce trust region by minimizing local quadratic model and recompute hookstep." << endl;
	  deltaMaxLocal = delta;
	  lambda = adjustLambda(lambda, lambdaMin, lambdaRequiredReduction, os);
	  delta  = adjustDelta(delta, lambda, deltaMin, deltaMax, os);

	  if (delta > snorm) {
	    if (os) *os << "That delta is still bigger than the Newton step." << endl;
	    //if (os) *os << "This shouldn't happen. Control structure needs review." << endl;
	    if (os) *os << "Reducing delta to half the length of the Newton step." << endl;
	    if (os) *os << "  old delta == " << delta << endl;
	    delta = 0.5*snorm;
	    if (os) *os << "  new delta == " << delta << endl;
	  }
	  //delta_has_decreased = true;
	  recompute_hookstep = true;
	}
      }

      // CASE 2: EXCELLENT IMPROVEMENT AND ROOM TO GROW
      // Improvement == Accurate or Negacurve means we're likely to get a significant improvement
      // by increasing trust region. Increase trust region by a fixed factor (rather than quadratic
      // model) so that trust-region search is monotonic.
      // Exceptions are
      //   have_backup && backup_rS < rS -- residual is getting wrose as we increase delta
      //   hookstep==newtonstep          -- increasing trust region won't change answer
      //   delta>=deltamax               -- we're already at the largest allowable trust region
      else if ((improvement == NegaCurve || improvement == Accurate)
	       && !(have_backup && backup_rS < rS)
	       && !hookstep_equals_newtonstep
	       && !(delta >= deltaMax)) {

	if (os) *os << "Improvement is " << improvement << endl;

	//lambda = adjustLambda(lambda, lambdaMin, lambdaMax);
	Real new_delta = adjustDelta(delta, lambdaMax, deltaMin, deltaMax, os);

	if (new_delta < deltaMaxLocal) {
	  if (os) *os << "Continue adjusting trust region radius delta because" << endl;
	  if (os) *os << "Suggested delta has room: new_delta < deltaMaxLocal " << new_delta << " < " << deltaMaxLocal << endl;
	  if (have_backup)
	    if (os) *os << "And residual is improving:     rS < backup_sS     " << rS << " < " << backup_rS << endl;

	  if (os) *os << "Increase delta and recompute hookstep." << endl;
	  have_backup = true;
	  backup_dx = dx;
	  backup_fS = fS;
	  backup_rS = rS;
	  backup_snorm = snorm;
	  backup_delta = delta;

	  recompute_hookstep = true;
	  if (os) *os << " old delta == " << delta << endl;
	  delta = new_delta;
	  if (os) *os << " new delta == " << delta << endl;
	}
	else {
	  if (os) *os << "Stop adjusting trust region radius delta and take step because the new delta" << endl;
	  if (os) *os << "reached a local limit:  new_delta >= deltaMaxLocal " << new_delta << " >= " << deltaMaxLocal << endl;
	  if (os) *os << "Reset delta to local limit and go to next Newton iteration" << endl;
	  delta = deltaMaxLocal;
	  if (os) *os << "  delta == " << delta << endl;
	  recompute_hookstep = false;
	}
      }

      // CASE 3: MODERATE IMPROVEMENT, NO ROOM TO GROW, OR BACKUP IS BETTER
      // Remaining cases: Improvement is acceptable: either Poor, Ok, Good, or
      // {Accurate/NegaCurve and (backup is superior || Hookstep==NewtonStep || delta>=deltaMaxLocal)}.
      // In all these cases take the current step or backup if better.
      // Adjust delta according to accuracy of quadratic prediction of residual (Poor, Ok, etc).
      else {
	if (os) *os << "Improvement is " << improvement <<  " (some form of acceptable)." << endl;
	if (os) *os << "Stop adjusting trust region and take a step because" << endl;
	if (os) *os << "Improvement is merely Poor, Ok, or Good : " << (improvement == Poor || improvement == Ok || improvement == Good) << endl;
	if (os) *os << "Newton step is within trust region      : " << hookstep_equals_newtonstep << endl;
	if (os) *os << "Backup step is better                   : " << (have_backup && backup_rS < rS) << endl;
	if (os) *os << "Delta has reached local limit           : " << (delta >= deltaMax) << endl;

	recompute_hookstep = false;

	// Backup step is better. Take it instead of current step.
	if (have_backup && backup_rS < rS) {
	  if (os) *os << "Take backup step and set delta to backup value." << endl;
	  dx = backup_dx;
	  fS = backup_fS;
	  rS = backup_rS;

	  if (os) *os << "  old delta == " << delta << endl;
	  snorm = backup_snorm;
	  delta = backup_delta;
	  if (os) *os << "  new delta == " << delta << endl;
	}

	// Current step is best available. Take it. Adjust delta according to Poor, Ok, etc.
	// Will start new monotonic trust-region search in Newton-hookstep, so be more
	// flexible about adjusting delta and use quadratic model.
	else {
	  if (have_backup) {
	    if (os) *os << "Take current step and keep current delta, since it's produced best result." << endl;
	    if (os) *os << "delta == " << delta << endl;
	    // Keep current delta because current step was arrived at through a sequence of
	    // adjustments in delta, and this value yielded best results.
	  }
	  else if (improvement == Poor) {
	    if (os) *os << "Take current step and reduce delta, because improvement is poor." << endl;
	    lambda = adjustLambda(lambda, lambdaMin, 1.0, os);
	    delta  = adjustDelta(delta, lambda, deltaMin, deltaMax, os);
	  }
	  else if (improvement == Ok || hookstep_equals_newtonstep || (delta >= deltaMax)) {
	    if (os) *os << "Take current step and leave delta unchanged, for the following reasons:" << endl;
	    if (os) *os << "improvement            : " << improvement << endl;
	    if (os) *os << "|newtonstep| <= delta  : " << (hookstep_equals_newtonstep ? "true" : "false") << endl;
	    if (os) *os << "delta >= deltaMax      : " << (delta >= deltaMax    ? "true" : "false") << endl;
	    //if (os) *os << "delta has decreased    : " << (delta_has_decreased ? "true" : "false") << endl;
	    if (os) *os << "delta == " << delta << endl;
	  }
	  else {  // improvement == Good, Accurate, or NegaCurve and no restriction on increasing apply
	    if (os) *os << "Take step and increase delta, if there's room." << endl;
	    lambda = adjustLambda(lambda, 1.0, lambdaMax, os);
	    delta  = adjustDelta(delta, lambda, deltaMin, deltaMax, os);
	  }
	}
      }
      //if (os) *os << "Recompute hookstep ? " << (recompute_hookstep ? "yes" : "no") << endl;

      if (recompute_hookstep)
	continue;
      else
	break;
    }

    if (os) *os << "Taking best step and continuing Newton iteration." <<endl;
    x += dx;
    dxnorm = L2Norm(dx);
    xnorm  = L2Norm(x);

    fx = fS;

    if (os) *os << "L2Norm(f(x)) == " << L2Norm(fx) << endl;
    if (os) *os << "rx           == " << rx << endl;
    if (os) *os << "L2Norm(sN)   == " << L2Norm(sN) << endl;
    if (os) *os << "L2Norm(sH)   == " << L2Norm(sH) << endl;
    if (os) *os << "L2Norm(dx)   == " << dxnorm << endl;

  }
  unscale(x,xscale);
  return x;
}

ostream& operator<<(ostream& os, HookstepPhase p) {
  if      (p == ReducingDelta)   os << "ReducingDelta";
  else if (p == IncreasingDelta) os << "IncreasingDelta";
  else                           os << "Finished";
  return os;
}
ostream& operator<<(ostream& os, ResidualImprovement i) {
  if      (i == Unacceptable) os << "Unacceptable";
  else if (i == Poor)         os << "Poor";
  else if (i == Ok)           os << "Ok";
  else if (i == Good)         os << "Good";
  else if (i == Accurate)     os << "Accurate";
  else                        os << "NegativeCurvature";
  return os;
 }

void rescale(ColumnVector& x, const ColumnVector& xscale) {
  if (xscale.length()  != 0)
    for (int i=0; i<x.length(); ++i)
      x(i) /= xscale(i);
}
void unscale(ColumnVector& x, const ColumnVector& xscale) {
  if (xscale.length()  != 0)
    for (int i=0; i<x.length(); ++i)
      x(i) *= xscale(i);
}

Real adjustDelta(Real delta, Real deltaRate, Real deltaMin, Real deltaMax, ostream* os) {
  if (os) *os  << "  old delta == " << delta << endl;

  // Check to see if we're already running with delta == deltaMin.
  // If so, and if deltaRate calls for furhter reduction, search has
  // ground to a halt. Error and exit.
  if (delta <= 1.00001*deltaMin && deltaRate < 1.0) {
    if (os) {
      *os << "    delta == " << delta << endl;
      *os << "delta new == " << delta*deltaRate << endl;
      *os << "delta min == " << deltaMin << endl;
      *os << "Sorry, can't go below deltaMin. Exiting." << endl;
    }
    exit(1);
  }

  delta *= deltaRate;

  if (delta < deltaMin) {
    delta = deltaMin;
    if (os) *os << "delta bottomed out at deltaMin" << endl;
  }
  else if (delta > deltaMax) {
    delta = deltaMax;
    if (os) *os << "delta topped out at deltaMax" << endl;
  }
  if (os) *os << "  new delta == " << delta << endl;
  return delta;
}

Real adjustLambda(Real lambda, Real lambdaMin, Real lambdaMax, ostream* os) {
  if (lambda < lambdaMin) {
    if (os) *os << "lambda == " << lambda << " is too small. Resetting to " << endl;
    lambda = lambdaMin;
    if (os) *os << "lambda == " << lambda << endl;
  }
  else if (lambda < lambdaMin) {
    if (os) *os << "lambda == " << lambda << " is too large. Resetting to " << endl;
    lambda = lambdaMax;
    if (os) *os << "lambda == " << lambda << endl;
  }
  return lambda;
}


void solve(const Matrix& Ut, const DiagMatrix& D, const Matrix& V, const ColumnVector& b,
	   ColumnVector& x) {

  ColumnVector bh = Ut*b;
  const Real eps = 1e-12;
  for (int i=0; i<D.rows(); ++i) {
    if (abs(D(i,i)) > eps)
      bh(i) *= 1.0/D(i,i);
    else
      bh(i)  = 0.0;
  }
  x = V*bh;
}

Real L2IP(const ColumnVector& u, const ColumnVector& v) {
  if (u.length() != v.length()) {
    cerr << "error in L2IP(ColumnVector, ColumnVector) : vector length mismatch" << endl;
    exit(1);
  }
  Real sum = 0.0;
  for (int i=0; i<u.length(); ++i)
    sum += u(i) * v(i);
  return sum;
}

// returns u dot v/(|u||v|) or zero if either has zero norm
Real align(const ColumnVector& u, const ColumnVector& v) {
  Real norm = L2Norm(u)*L2Norm(v);
  if (norm == 0.0)
    return 0.0;
  else
    return L2IP(u,v)/norm;
}

Real bisectSearch(R2Rfunc f, Real a, Real b, Real feps, int maxsteps) {
  Real fa = (*f)(a);
  Real fb = (*f)(b);
  if (fa*fb > 0)
    cferror("bisectSearch(a,b,fn,xn) : a and b don't bracket a zero");

  Real c = 0.5*(a+b)/2;
  Real fc;

  for (int n=0; n<maxsteps; ++n) {
    c = 0.5*(a+b);
    fc = (*f)(c);
    if (abs(fc) < feps)
      break;
    if (fc*fa > 0) {
      a = c;
      fa = fc;
    }
    else {
      b = c;
      fb = fc;
    }
  }
  return c;
}


Real GMRESHookstep(FlowField& u, Real& T, FieldSymmetry& sigma, Real Reynolds, const GMRESHookstepFlags& searchflags, const DNSFlags& dnsflags, const TimeStep& dtarg) {

  // get short names for the following heavily used searchflags fields
  ostream* os = searchflags.logstream; // a short name for ease of use

  Real delta  = searchflags.delta;

  //const BasisType nrm = l2basis ? L2Basis : AdHocBasis;
  const BasisType nrm = AdHocBasis;
  //const int w = coutprec + 7;
  const bool Tsearch  = searchflags.orbit();
  const bool Rsearch  = searchflags.relative();
  const bool Tnormalize = Tsearch ? false : true;

  TimeStep dt(dtarg);
  int fcount_newton = 0;
  int fcount_hookstep = 0;

  const int Nx = u.Nx();
  const int Ny = u.Ny();
  const int Nz = u.Nz();
  const Real Lx = u.Lx();
  const Real ya = u.a();
  const Real yb = u.b();
  const Real Lz = u.Lz();

  //const int kxmax = u.kxmaxDealiased();
  //const int kzmax = u.kzmaxDealiased();

  vector<RealProfileNG> basis;
  /**********************
  if (l2basis || s1 || s2 || s3) {
    os << "constructing basis set (takes a long time)..." << endl;
    basis = realBasisNG( Ny, kxmax, kzmax, Lx, Lz, ya, yb);

    vector<FieldSymmetry> symms;
    if (s1) symms.push_back(FieldSymmetry( 1, 1,-1, 0.5, 0.0));
    if (s2) symms.push_back(FieldSymmetry(-1,-1, 1, 0.5, 0.5));
    if (s3) symms.push_back(FieldSymmetry(-1,-1,-1, 0.0, 0.5));
    if (s1 || s2 || s3)
      selectSymmetries(basis, symms);
    orthonormalize(basis);
  }
  **************************/

  project(dnsflags.symmetries, u, "initial guess u", *os);
  fixdivnoslip(u);

  ColumnVector dxN;   // "dx Newton" : the Newton step, computed w GMRES
  field2vector(u, dxN, basis);

  // Mcon == number of constraints
  // Neqn == number of equations (total)
  // Nunk == number of unknowns

  const int uunk = dxN.length();                                  // # of variables for u unknonwn
  const int Tunk = Tsearch ?  uunk : -1;                         // index for T unknown
  const int xunk = Rsearch ? (uunk + (Tsearch ? 1 : 0)) : -1;    // index for x-shift unknown
  const int zunk = Rsearch ? (xunk + 1) : -1;                    // index for z-shift unknown
  const int Nunk = uunk + (Tsearch ? 1 : 0) + (Rsearch ? 2 : 0);
  const int xeqn = xunk;
  const int zeqn = zunk;
  const int Teqn = Tunk;
  const int Neqn = Nunk;
  const int Ngmres2 = (Nunk < searchflags.Ngmres) ? Nunk : searchflags.Ngmres;


  if (os) {
    *os << Neqn << " equations" << endl;
    *os << Nunk << " unknowns" << endl;
    if (Tsearch) {
      *os << "Teqn  == " << Teqn << endl;
      *os << "Tunk  == " << Tunk << endl;
    }
    if (Rsearch) {
      *os << "xeqn  == " << xeqn << endl;
      *os << "xunk  == " << xunk << endl;
      *os << "zeqn  == " << zeqn << endl;
      *os << "zunk  == " << zunk << endl;
    }
  }

  dxN = ColumnVector(Nunk);

  // These are purely diagnostic variables. I want to see how these directions change per iteration
  //ColumnVector prev_dxG(Nunk); // previous gradient direction
  ColumnVector prev_dxH(Nunk); // previous hookstep
  ColumnVector prev_dxN(Nunk); // previous newton

  // ==============================================================
  // Initialize Newton iteration
  // Notation:
  //    x = set of free variables: (u), (u,T), (u,sigma), or (u,sigma,T)
  // N means Newton, H means Hookstep, e.g.
  //   dxN is the Newton step
  //   dxH is the Hookstep

  FlowField      du(Nx, Ny, Nz, 3, Lx, Lz, ya, yb); // temporary for increments in orbit velocity field
  FlowField      Gx(Nx, Ny, Nz, 3, Lx, Lz, ya, yb); // G(x) = G(u,sigma,T) = sigma f^T(u) - u
  FlowField   DG_dx(Nx, Ny, Nz, 3, Lx, Lz, ya, yb); // "DG*dx" = 1/e [G(u+e.du, sigma+e.dsigma, T+e.dT) - G(u,sigma,T)]
  FlowField       p(Nx, Ny, Nz, 1, Lx, Lz, ya, yb); // pressure field for time integration
  FlowField      uH(Nx, Ny, Nz, 3, Lx, Lz, ya, yb); // stores Newton uN = u + duN and Hookstep uH = u + duH fields
  FlowField      GH(Nx, Ny, Nz, 3, Lx, Lz, ya, yb); // stores GN = G(dx+dxH) and GH = G(dx + dxH)

  Real tH  = 0.0;  // stores tN = T + dTN and tH = T + dTH
  Real dT  = 0.0;  // Newton step for orbit period T
  Real dax = 0.0;                                    // Newton step for x phase shift
  Real daz = 0.0;                                    // Newton step for z phase shift

  //FieldSymmetry sigma(sx,sy,sz,ax,az,anti);
  //if (symname.length() != 0)
  //load(symname, T, sigma);
  *os << "Initial integration time T == " << T << endl;
  *os << "Initial symmetry sigma     == " << sigma << endl;

  FieldSymmetry dsigma; // identity, for now
  FieldSymmetry sigmaH(sigma);

  ChebyCoeff U(Ny, ya, yb, Spectral);
  U[1] = 1;

  Real rx = 0.0;        // Dennis & Schnabel residual r(u,T) = 1/2 ||G(x)||^2 == 1/2 VNorm2(G(x))
  Real gx = 0.0;        // g(u,T) = L2Norm(G(x))
  Real init_rx = 0.0;
  Real init_gx = 0.0;
  Real prev_rx = 0.0;
  Real prev_gx = 0.0;
  Real gmres_residual = 0.0;
  Real dxHnorm  = 0.0;   // norm of hookstep
  Real dxNnorm  = 0.0;   // norm of newton step
  Real dunorm  = 0.0;
  Real unorm   = 0.0;
  Real CFL = 0.0;
  bool stuck = false;

  ofstream osreynolds((searchflags.outdir + "Reynolds.asc").c_str());
  osreynolds << setprecision(17) << Reynolds << endl;
  osreynolds.close();

  ofstream osconv((searchflags.outdir + "convergence.asc").c_str());
  osconv.setf(ios::left);
  osconv << setw(14) << "% L2Norm(G)" << setw(14) << "r(u,T)" << setw(14) << "delta";
  if (Tsearch)
    osconv << setw(14) << "dT";
  if (Rsearch) {
    osconv << setw(14) << "dax";
    osconv << setw(14) << "daz";
  }
  osconv << setw(14) << "L2Norm(du)" << setw(14) << "L2Norm(dxN)" << setw(14) << "L2Norm(dxH)"
	 << setw(14) << "|DG dxN + G(x)|/|G(x)|" << setw(10) << "ftotal" << setw(10) << "fnewt"
	 << setw(10) << "fhook" << endl;

  // Determine which of four half-cell translations sigma minimizes 1/2 || sigma f^T(u) - u ||^2
  //*os << "Determining which of four half-cell translations sigma minimizes residual" << endl;
  if (os) {
    *os << "Search for solutions of sigma f^T(u) - u == 0 with " << endl;
    *os << "    sigma == " << sigma << endl;
    *os << " Reynolds == " << Reynolds << endl;
    *os << setprecision(6);
    *os << "Computing G(x) = " << ((Tnormalize) ? "1/T (sigma f(u,T) - u)" : "sigma f(u,T) - u") << endl;
  }

  G(u, T, sigma, Gx, Reynolds, dnsflags, dt, Tnormalize, fcount_hookstep, CFL, os);
  if (os) *os << endl;

  for (int newtonStep = 0; newtonStep<=searchflags.Nnewton; ++newtonStep) {

    if (os) {
      *os << "========================================================" << endl;
      *os << "Newton iteration number " << newtonStep << endl;
      *os << "Computing normalized dudt, edudx, edudz" << endl;
    }

    // Compute quantities that will be used multiple times in GMRES loop
    // Gx     = (sigma f^T(u) - u)  or  1/T (sigma f^T(u) - u)
    // dfxdT  = 1/(dt) (f^(T+dt)(u) - f^(T)(u))
    // dudt   = 1/(dt) (f^(dt)(u) - u)
    //        = tangent vector to flow
    // edudx  = tangent vector of x translation
    // edudz  = tangent vector of z translation

    dunorm = L2Norm(du);
    unorm = L2Norm(u);
    rx = 0.5*Norm2(Gx, nrm);
    gx = L2Norm(Gx);

    if (newtonStep == 0) {
      init_gx = gx;
      prev_gx = gx;
      init_rx = rx;
      prev_rx = rx;
    }

    FlowField uU(u);
    uU += U;

    if (os) {
      *os << "Current state of Newton iteration:" << endl;
      *os << "    L2Norm(u)  == " << unorm << endl;
      *os << "   L2Norm(du)  == " << dunorm <<endl;
      *os << "  L2Norm(dxH)  == " << dxHnorm <<endl;
      *os << "           T   == " << T << endl;
      *os << "          dT   == " << dT << endl;
      *os << "         sigma == " << sigma << endl;
      *os << "        dsigma == " << dsigma << endl;
      //*os << "L2Dist(u,u0) == " << L2Dist(u, FlowField(uname)) << endl;
      *os << "gx == L2Norm(G(x)) : " << endl;
      *os << "   initial  gx == " << init_gx << endl;
      *os << "   previous gx == " << prev_gx << endl;
      *os << "   current  gx == " << gx << endl;
      *os << "rx == 1/2 Norm2(G(x)) : " << endl;
      *os << "   initial  rx == " << init_rx << endl;
      *os << "   previous rx == " << prev_rx << endl;
      *os << "   current  rx == " << rx << endl;
      *os << "1/2 L2Norm2(G,u,T) : " << endl;
      *os << "               == " << 0.5*square(gx) << endl;
      *os << "CFL == " << CFL << endl;
    }
    osconv.setf(ios::left); osconv << setw(14) << gx;
    osconv.setf(ios::left); osconv << setw(14) << rx;
    osconv.setf(ios::left); osconv << setw(14) << delta;
    if (Tsearch) {
      osconv.setf(ios::left); osconv << setw(14) << dT;
    }
    if (Rsearch) {
      osconv.setf(ios::left); osconv << setw(14) << dsigma.ax();
      osconv.setf(ios::left); osconv << setw(14) << dsigma.az();
    }
    osconv.setf(ios::left); osconv << setw(14) << dunorm;
    osconv.setf(ios::left); osconv << setw(14) << dxNnorm;
    osconv.setf(ios::left); osconv << setw(14) << dxHnorm;
    osconv.setf(ios::left); osconv << setw(14) << gmres_residual;
    osconv.setf(ios::left); osconv << setw(10) << fcount_newton + fcount_hookstep;
    osconv.setf(ios::left); osconv << setw(10) << fcount_newton;
    osconv.setf(ios::left); osconv << setw(10) << fcount_hookstep;
    osconv << endl;

    string ns = i2s(newtonStep);
    u.save(searchflags.outdir + "ubest");
    sigma.save(searchflags.outdir + "sigmabest");
    save(T, searchflags.outdir + "Tbest");
    save(Reynolds, searchflags.outdir + "Reynolds");

    if (searchflags.saveall) {
      u.save(searchflags.outdir + "unewt" + ns);
      if (Tsearch || Rsearch) {
	sigma.save(searchflags.outdir + "sigmanewt" + ns);
	save(T, searchflags.outdir + "Tnewt" + ns);
      }
    }

    if (gx < searchflags.epsSearch) {
      if (os) *os << "Newton search converged. Returning residual" << endl
		 << "L2Norm(G(x))) == " << gx << " < "
		 << searchflags.epsSearch << " == " << searchflags.epsSearch << endl;
      return gx;
    }
    else if (stuck) {
      if (os) *os << "Newton search is stuck. Returning best answer so far." << endl
		 << "L2Norm(G(x))) == " << gx << endl;
      return gx;
    }
    else if (newtonStep == searchflags.Nnewton) {
      if (os) *os << "Reached maximum number of Newton steps. Returning best answer so far." << endl
		 << "L2Norm(G(x))) == " << gx << endl;
      return gx;
    }

    prev_gx = gx;
    prev_rx = rx;

    // Calculate three unit vectors to which du must be orthogonal
    FlowField edudx = xdiff(u);
    FlowField edudz = zdiff(u);
    FlowField edudt(u);
    f(u, 1, searchflags.epsDt, edudt, Reynolds, dnsflags, os);   // using dt=epsDt timestep
    edudt -= u;
    edudt *= 1.0/searchflags.epsDt;

    // Set Tscale so that |dT|*Tscale = TscaleRel |du|
    // Tscale = TscaleRel |du|/|dt|
    // O(t)*Tscale = O(u) => Tscale = u/t
    //const Real Tscale = TscaleRel*V2Norm(edudt);
    const Real Tscale = searchflags.TscaleRel;
    if (os) *os << "delta  == " << delta << endl;
    if (os) *os << "Tscale == " << Tscale << endl;
    edudt *= 1.0/L2Norm(edudt);
    edudx *= 1.0/L2Norm(edudx);
    edudz *= 1.0/L2Norm(edudz);

    // Set up RHS vector b = (-field2vector(G(x)), 0, 0, 0);
    // (0, 0, 0) is the RHS of eqn enforcing orthog. to (edudt, edudx, edudz)
    ColumnVector b(Neqn);
    setToZero(b);
    field2vector(Gx, b, basis);
    if (Tsearch)
      b(Teqn) = 0;
    if (Rsearch) {
      b(xeqn) = 0;
      b(zeqn) = 0;
    }
    b *= -1.0;

    GMRES gmres(b, Ngmres2, searchflags.epsKrylov);

    // ===============================================================
    // GMRES iteration to solve DG(x) dx = -G(x) for dx = (du, dsigma, dT)
    for (int n=0; n<Ngmres2; ++n) {

      if (os) *os << "Newt,GMRES == " << newtonStep << ',' << n << ", " << flush;

      // Compute v = Ab in Arnoldi iteration terms, where b is Q.column(n)
      // In Navier-Stokes terms, the main quantity to compute is
      // DG dx = 1/e (G(u + e du, sigma + e dsigma, T + e dT) - G(u,sigma,T)) for e << 1

      ColumnVector q = gmres.testVector();
      vector2field(q, du, basis);
      dT  = Tsearch ? q(Tunk)*Tscale : 0.0;
      dax = Rsearch ? q(xunk)*searchflags.Rscale : 0.0;
      daz = Rsearch ? q(zunk)*searchflags.Rscale : 0.0;
      dsigma = FieldSymmetry(1,1,1,dax,daz);

      // Compute DG dx = 1/e (G(u+e.du, sigma+e.dsigma, T+e.dT) - G(u,sigma,T))
      DG(u, du, T, dT, sigma, dsigma, Gx, DG_dx, Reynolds, dnsflags, dt, Tnormalize,
	 searchflags.epsDu,  searchflags.centdiff, fcount_newton, CFL, os);

      ColumnVector Lq(Neqn);
      field2vector(DG_dx, Lq, basis);
      if (Tsearch)
	Lq(Teqn) = searchflags.dudtortho ? searchflags.orthoscale*L2InnerProduct(du, edudt) : 0.0;
      if (Rsearch) {
	Lq(xeqn) = searchflags.dudxortho ? searchflags.orthoscale*L2InnerProduct(du, edudx) : 0.0;
	Lq(zeqn) = searchflags.dudxortho ? searchflags.orthoscale*L2InnerProduct(du, edudz) : 0.0;
      }

      gmres.iterate(Lq);
      gmres_residual = gmres.residual();

      if (os) *os << " res == " << gmres_residual << endl;

      if (gmres_residual < searchflags.epsGMRES) {
        if (os) *os << "GMRES converged. Breaking." <<endl;
	dxN = gmres.solution();
	break;
      }
      else if (n==searchflags.Ngmres-1 && gmres_residual < searchflags.epsGMRESf ) {
        if (os) *os << "GMRES has not converged, but the final iterate is acceptable. Breaking." <<endl;
	dxN = gmres.solution();
      }
      else if (n==searchflags.Ngmres-1) {
	if (os) *os << "GMRES failed to converge. Returning best answer so far." << endl;
	return gx;
      }
    } // end GMRES iteration

    dxNnorm = L2Norm(dxN);

    // Diagnostic
    //vector2field(dxN, du, basis);
    //du.save("dxN" + i2s(newtonStep));

    // ==================================================================
    // Hookstep algorithm

    if (os) *os << "------------------------------------------------" << endl;
    if (os) *os << "Beginning hookstep calculations." << endl;

    int Nk = gmres.n(); // Krylov dimension
    Matrix Hn = gmres.Hn();
    Matrix Qn = gmres.Qn();
    Matrix Qn1t = gmres.Qn1().transpose(); // wasteful of memory
    SVD svd(Hn, SVD::std);

    Matrix V = svd.right_singular_matrix();
    Matrix Ut = svd.left_singular_matrix().transpose();
    DiagMatrix D = svd.singular_values();

    ColumnVector btmp = Qn1t*b;
    ColumnVector bh = Ut*btmp; // b hat

    int hookcount = 0;

    // Hookstep algorithm requires iterated tweaking. Remember the best hookstep
    // with these variables, so we can revert to it if necessary.
    bool have_backup  = false;
    Real backup_dxHnorm = 0.0;
    Real backup_delta = 0.0;
    Real backup_rH    = 0.0;
    Real backup_dT    = 0;
    FlowField backup_du;
    FlowField backup_GH;
    FieldSymmetry backup_dsigma;

    ColumnVector dxH;  // The hookstep
    //ColumnVector dxG;  // The gradient direction

    //bool delta_has_decreased = false;
    Real deltaMaxLocal = searchflags.deltaMax;

    // Find a good trust region and the optimal hookstep in it.
    while (true) {

      if (os) *os << "-------------------------------------------" << endl;
      if (os) *os << "Newton, hookstep number " << newtonStep << ", "
	   << hookcount++ << endl;
      if (os) *os << "delta == " << delta << endl;

      bool hookstep_equals_newtonstep;

      if (L2Norm(dxN) <= delta) {
	if (os) *os << "Newton step is within trust region: " << endl;
	if (os) *os << "L2Norm(dxN) == " << L2Norm(dxN) << " <= " << delta << " == delta" << endl;
	hookstep_equals_newtonstep = true;
	dxH = dxN;
      }
      else {
	if (os) *os << "Newton step is outside trust region: " << endl;
	if (os) *os << "L2Norm(dxN) == " << L2Norm(dxN) << " > " << delta << " == delta" << endl;
	if (os) *os << "Calculate hookstep dxH(mu) with radius L2Norm(dxH(mu)) == delta" << endl;

	hookstep_equals_newtonstep = false;

	// This for-loop determines the hookstep. Search for value of mu that
	// produces ||dxHmu|| == delta. That provides optimal reduction of
	// quadratic model of residual in trust region norm(dxHmu) <= delta.
	ColumnVector dxHmu(Nk); // dxH(mu) : Hookstep dxH as a function of parameter
	Real ndxHmu = 0;        // norm of dxHmu
	Real mu = 0;            // a parameter we search over to find dxH with radius delta

	// See Dennis and Schnabel for this search-over-mu algorithm
	for (int hookSearch=0; hookSearch<searchflags.Nhook; ++hookSearch) {
	  for (int i=0; i<Nk; ++i)
	    dxHmu(i) = bh(i)/(D(i,i) + mu); // research notes
	  ndxHmu = L2Norm(dxHmu);
	  Real Phi = ndxHmu*ndxHmu - delta*delta;
	  if (os) *os << "mu, L2Norm(dxH(mu)) == " << mu << ", " << ndxHmu << endl;

	  // Found satisfactory value of mu and thus dxHmu and dxH s.t. |dxH| < delta
	  if (ndxHmu < delta || (ndxHmu > (1-searchflags.deltaFuzz)*delta && ndxHmu < (1+searchflags.deltaFuzz)*delta))
	    break;

	  // Update value of mu and try again. Update rule is a Newton search for
	  // Phi(mu)==0 based on a model of form a/(b+mu) for norm(sh(mu)). See
	  // Dennis & Schnabel.
	  else if (hookSearch<searchflags.Nhook-1) {
	    Real PhiPrime = 0.0;
	    for (int i=0; i<Nk; ++i) {
	      Real di_mu = D(i,i) + mu;
	      Real bi = bh(i);
	      PhiPrime -= 2*bi*bi/(di_mu*di_mu*di_mu);
	    }
	    mu -= (ndxHmu/delta) * (Phi/PhiPrime);
	  }
	  else {
	    if (os) *os << "Couldn't find solution of hookstep optimization eqn Phi(mu)==0" << endl;
	    if (os) *os << "This shouldn't happen. It indicates an error in the algorithm." << endl;
	    if (os) *os << "Returning best answer so far\n";
	    return gx;
	  }
	} // search over mu for norm(s) == delta

	if (os) *os << "Found hookstep of proper radius" << endl;
	if (os) *os << "ndxHmu == " << L2Norm(dxHmu) << " ~= " << delta << " == delta " << endl;
	dxH = Qn*(V*dxHmu);
      } // end else clause for Newton step outside trust region


      // Compute
      // (1) actual residual from evaluation of
      //     rH == r(x+dxH) act  == 1/2 (G(x+dxH), G(x+dxH))
      //
      // (2) predicted residual from quadratic model of r(x)
      //     rP == r(x+dx) pred  == 1/2 (G(x), G(x)) + (G(x), DG dx)
      //
      // (3) slope of r(x)
      //     dr/dx == (r(x + dx) - r(x))/|dx|
      //           == (G(x), DG dx)/|dx|
      // where ( , ) is FlowField inner product V2IP, which equals L2 norm of vector rep

      // (1) Compute actual residual of step, rH
      if (os) *os << "Computing residual of hookstep dxH" << endl;
      vector2field(dxH, du, basis);
      project(dnsflags.symmetries, du, "Newton step du", *os);
      fixdivnoslip(du);

      dunorm = L2Norm(du);
      uH  = u;
      uH += du;
      dT  = Tsearch ? dxH(Tunk)*Tscale : 0;
      dax = Rsearch ? dxH(xunk)*searchflags.Rscale : 0.0;
      daz = Rsearch ? dxH(zunk)*searchflags.Rscale : 0.0;
      dsigma = FieldSymmetry(1,1,1,dax,daz);

      if (os) *os << "L2Norm(du) == " << L2Norm(du) << endl;
      if (os) *os << "       dT  == " << dT << endl;
      if (os) *os << "   dsigma  == " << dsigma << endl;
      tH  = T + dT;
      sigmaH = dsigma*sigma;
      G(uH, tH, sigmaH, GH, Reynolds, dnsflags, dt, Tnormalize, fcount_hookstep, CFL, os);
      if (os) *os << endl;

      Real rH = 0.5*Norm2(GH, nrm); // actual residual of hookstep
      Real Delta_rH = rH - rx;     // improvement in residual
      if (os) *os << "r(x), r(x+dxH) == " << rx << ", " << rH << endl;

      // (2) and (3) Compute quadratic model and slope
      if (os) *os << "Computing local quadratic model of residual" << endl;
      DG(u, du, T, dT, sigma, dsigma, Gx, DG_dx, Reynolds, dnsflags, dt, Tnormalize,
	 searchflags.epsDu,  searchflags.centdiff, fcount_newton, CFL, os);
      if (os) *os << endl;

      // Local quadratic and linear models of residual, based on Taylor exp at current position
      // Quadratic model of r(x+dx): rQ == 1/2 |G(x) + DG dx|^2
      // Linear  model   of r(x+dx): rL == 1/2 (G(x), G(x)) + (G(x), DG dx)

      Real Delta_rL = IP(Gx,DG_dx,nrm); // rL - 1/2 (G(x), G(x)) == (G(x), DG dx)
      Real rL = rx + Delta_rL;          // rL == 1/2 (G(x), G(x)) + (G(x), DG dx)

      if (rL >= rx) {
	if (os) *os << "error : local linear model of residual is increasing, indicating\n"
	     << "        that the solution to the Newton equations is inaccurate\n";

	if (searchflags.centdiff == true) {
	  if (os) *os << "Returnign best answer so far." << endl;
	  return gx;
	}
	else {
	  if (os) *os << "Trying local linear model again, using centered finite differencing" << endl;
	  DG(u, du, T, dT, sigma, dsigma, Gx, DG_dx, Reynolds, dnsflags, dt, Tnormalize,
	     searchflags.epsDu,  true, fcount_newton, CFL, os);

	  Delta_rL = IP(Gx,DG_dx,nrm); // rL - 1/2 (G(x),G(x)) == (G(x), DG dx)
	  rL = rx + Delta_rL;          // rL == 1/2 (G(x),G(x)) + (G(x), DG dx)

	  if (rL >= rx) {
	    if (os) *os << "error : centered differencing didn't help\n"
		 << "Returning best answer so far." << endl;
	    return gx;
	  }
	  // if we get here, centered differencing does help, and we can continue
	}
      }

      DG_dx += Gx;
      Real rQ = 0.5*Norm2(DG_dx,nrm);   // rQ == 1/2 |G(x) + DG dx|^2
      Real Delta_rQ = rQ - rx;          // rQ == 1/2 (G(x),G(x)) - 1/2 |G(x) + DG dx|^2

      dxHnorm = L2Norm(dxH);

      // ===========================================================================
      if (os) *os << "Determining what to do with current Newton/hookstep and trust region" << endl;

      // Coefficients for non-local quadratic model of residual, based on r(x), r'(x) and r(x+dx)
      // dr/dx == (G(x), DG dx)/|dx|
      Real drdx  = Delta_rL/dxHnorm;

      // Try to minimize a quadratic model of the residual
      // r(x+dx) = r(x) + r' |dx| + 1/2 r'' |dx|^2
      Real lambda = -0.5*drdx*dxHnorm/(rH - rx - drdx*dxHnorm);

      // Compare the actual reduction in residual to the quadratic model of residual.
      // How well the model matches the actual reduction will determine, later on, how and
      // when the radius of the trust region should be changed.
      Real Delta_rH_req  = searchflags.improvReq*drdx*dxHnorm;  // the minimum acceptable change in residual
      Real Delta_rH_ok   = searchflags.improvOk*Delta_rQ;       // acceptable, but reduce trust region for next newton step
      Real Delta_rH_good = searchflags.improvGood*Delta_rQ;     // acceptable, keep same trust region in next newton step
      Real Delta_rQ_acc  = abs(searchflags.improvAcc*Delta_rQ); // for accurate models, increase trust region and recompute hookstep
      Real Delta_rH_accP = Delta_rQ + Delta_rQ_acc; //   upper bound for accurate predictions
      Real Delta_rH_accM = Delta_rQ - Delta_rQ_acc; //   lower bound for accurate predictions

      // Characterise change in residual Delta_rH
      ResidualImprovement improvement;  // fine-grained characterization

      // Place improvement in contiguous spectrum: Unacceptable > Poor > Ok > Good.
      if (Delta_rH > Delta_rH_req)
	improvement = Unacceptable;     // not even a tiny fraction of linear rediction
      else if (Delta_rH > Delta_rH_ok)
	improvement = Poor;             // worse than small fraction of quadratic prediction
      else if (Delta_rH < Delta_rH_ok && Delta_rH > Delta_rH_good)
	improvement = Ok;
      else {
	improvement = Good;             // not much worse or better than large fraction of prediction
	if (Delta_rH_accM <= Delta_rH &&  Delta_rH <= Delta_rH_accP)
	  improvement = Accurate;         // close to quadratic prediction
	else if (Delta_rH < Delta_rL)
	  improvement = NegaCurve;        // negative curvature in r(|s|) => try bigger step
      }
      const int w = 13;
      if (os) *os << "rx       == " << setw(w) << rx            << " residual at current position" << endl;
      if (os) *os << "rH       == " << setw(w) << rH            << " residual of newton/hookstep" << endl;
      if (os) *os << "Delta_rH == " << setw(w) << Delta_rH      << " actual improvement in residual from newton/hookstep" << endl;
      if (os) *os << endl;

      if (improvement==Unacceptable)
      if (os) *os << "Delta_rH == " << setw(w) << Delta_rH      << " ------> Unacceptable <------ " << endl;
      if (os) *os << "            " << setw(w) << Delta_rH_req  << " lower bound for acceptable improvement" << endl;
      if (improvement==Poor)
      if (os) *os << "Delta_rH == " << setw(w) << Delta_rH      << " ------> Poor <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rH_ok   << " upper bound for ok improvement." << endl;
      if (improvement==Ok)
      if (os) *os << "Delta_rH == " << setw(w) << Delta_rH      << " ------> Ok <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rH_good << " upper bound for good improvement." << endl;
      if (improvement==Good)
      if (os) *os << "Delta_rH == " << setw(w) << Delta_rH      << " ------> Good <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rH_accP << " upper bound for accurate prediction." << endl;
      if (improvement==Accurate)
      if (os) *os << "Delta_rH == " << setw(w) << Delta_rH      << " ------> Accurate <------" << endl;
      if (os) *os << "            " << setw(w) << Delta_rH_accM << " lower bound for accurate prediction." << endl;
      if (os) *os << "            " << setw(w) << Delta_rL      << " local linear model of improvement" << endl;
      if (improvement==NegaCurve)
      if (os) *os << "Delta_rH == " << setw(w) << Delta_rH      << " ------> NegativeCurvature <------" << endl;

      bool recompute_hookstep = false;

      if (os) *os << "lambda       == " << lambda << " is the reduction/increase factor for delta suggested by quadratic model" << endl;
      if (os) *os << "lambda*delta == " << lambda*delta << " is the delta suggested by quadratic model" << endl;

      // See comments at end for outline of this control structure

      // Following Dennis and Schnable, if the Newton step lies within the trust region,
      // reset trust region radius to the length of the Newton step, and then adjust it
      // further according to the quality of the residual.
      //if (hookstep_equals_newtonstep)
      //delta = dxHnorm;

      // CASE 1: UNACCEPTABLE IMPROVEMENT
      // Improvement is so bad (relative to gradient at current position) that we'll
      // in all likelihood get better results in a smaller trust region.
      if (improvement == Unacceptable) {
	if (os) *os << "Improvement is unacceptable." << endl;

	if (have_backup) {
	  if (os) *os << "But we have a backup step that was acceptable." << endl;
	  if (os) *os << "Revert to backup step and backup delta, take the step, and go to next Newton-GMRES iteration." << endl;
	  dsigma = backup_dsigma;
	  dT = backup_dT;
	  du = backup_du;
	  GH = backup_GH;
	  rH = backup_rH;
	  dxHnorm = backup_dxHnorm;
	  delta = backup_delta;

	  recompute_hookstep = false;
	}
	else {
	  if (os) *os << "No backup step is available." << endl;
	  if (os) *os << "Reduce trust region by minimizing local quadratic model and recompute hookstep." << endl;
	  deltaMaxLocal = delta;
	  lambda = adjustLambda(lambda, searchflags.lambdaMin, searchflags.lambdaRequiredReduction);
	  delta  = adjustDelta(delta, lambda, searchflags.deltaMin, searchflags.deltaMax, os);

	  if (delta > dxHnorm) {
	    if (os) *os << "That delta is still bigger than the Newton step." << endl;
	    //if (os) *os << "This shouldn't happen. Control structure needs review." << endl;
	    if (os) *os << "Reducing delta to half the length of the Newton step." << endl;
	    if (os) *os << "  old delta == " << delta << endl;
	    delta = 0.5*dxHnorm;
	    if (os) *os << "  new delta == " << delta << endl;
	  }
	  //delta_has_decreased = true;
	  recompute_hookstep = true;
	}
      }

      // CASE 2: EXCELLENT IMPROVEMENT AND ROOM TO GROW
      // Improvement == Accurate or Negacurve means we're likely to get a significant improvement
      // by increasing trust region. Increase trust region by a fixed factor (rather than quadratic
      // model) so that trust-region search is monotonic.
      // Exceptions are
      //   have_backup && backup_rH < rH -- residual is getting wrose as we increase delta
      //   hookstep==newtonstep          -- increasing trust region won't change answer
      //   delta>=deltamax               -- we're already at the largest allowable trust region
      else if ((improvement == NegaCurve || improvement == Accurate)
	       && !(have_backup && backup_rH < rH)
	       && !hookstep_equals_newtonstep
	       && !(delta >= searchflags.deltaMax)) {

	if (os) *os << "Improvement is " << improvement << endl;

	//lambda = adjustLambda(lambda, searchflags.lambdaMin, searchflags.lambdaMax);
	Real new_delta = adjustDelta(delta, searchflags.lambdaMax, searchflags.deltaMin, searchflags.deltaMax, os);

	if (new_delta < deltaMaxLocal) {
	  if (os) *os << "Continue adjusting trust region radius delta because" << endl;
	  if (os) *os << "Suggested delta has room: new_delta < deltaMaxLocal " << new_delta << " < " << deltaMaxLocal << endl;
	  if (have_backup)
	    if (os) *os << "And residual is improving:     rH < backup_sS     " << rH << " < " << backup_rH << endl;

	  if (os) *os << "Increase delta and recompute hookstep." << endl;
	  have_backup = true;
	  backup_dsigma = dsigma;
	  backup_dT = dT;
	  backup_du = du;
	  backup_GH = GH;
	  backup_rH = rH;
	  backup_dxHnorm = dxHnorm;
	  backup_delta = delta;

	  recompute_hookstep = true;
	  if (os) *os << " old delta == " << delta << endl;
	  delta = new_delta;
	  if (os) *os << " new delta == " << delta << endl;
	}
	else {
	  if (os) *os << "Stop adjusting trust region radius delta and take step because the new delta" << endl;
	  if (os) *os << "reached a local limit:  new_delta >= deltaMaxLocal " << new_delta << " >= " << deltaMaxLocal << endl;
	  if (os) *os << "Reset delta to local limit and go to next Newton iteration" << endl;
	  delta = deltaMaxLocal;
	  if (os) *os << "  delta == " << delta << endl;
	  recompute_hookstep = false;
	}
      }

      // CASE 3: MODERATE IMPROVEMENT, NO ROOM TO GROW, OR BACKUP IS BETTER
      // Remaining cases: Improvement is acceptable: either Poor, Ok, Good, or
      // {Accurate/NegaCurve and (backup is superior || Hookstep==NewtonStep || delta>=deltaMaxLocal)}.
      // In all these cases take the current step or backup if better.
      // Adjust delta according to accuracy of quadratic prediction of residual (Poor, Ok, etc).
      else {
	if (os) *os << "Improvement is " << improvement <<  " (some form of acceptable)." << endl;
	if (os) *os << "Stop adjusting trust region and take a step because" << endl;
	if (os) *os << "Improvement is merely Poor, Ok, or Good : " << (improvement == Poor || improvement == Ok || improvement == Good) << endl;
	if (os) *os << "Newton step is within trust region      : " << hookstep_equals_newtonstep << endl;
	if (os) *os << "Backup step is better                   : " << (have_backup && backup_rH < rH) << endl;
	if (os) *os << "Delta has reached local limit           : " << (delta >= searchflags.deltaMax) << endl;

	recompute_hookstep = false;

	// Backup step is better. Take it instead of current step.
	if (have_backup && backup_rH < rH) {
	  if (os) *os << "Take backup step and set delta to backup value." << endl;
	  dsigma = backup_dsigma;
	  dT = backup_dT;
	  du = backup_du;
	  GH = backup_GH;
	  rH = backup_rH;

	  if (os) *os << "  old delta == " << delta << endl;
	  dxHnorm = backup_dxHnorm;
	  delta = backup_delta;
	  if (os) *os << "  new delta == " << delta << endl;
	}

	// Current step is best available. Take it. Adjust delta according to Poor, Ok, etc.
	// Will start new monotonic trust-region search in Newton-hookstep, so be more
	// flexible about adjusting delta and use quadratic model.
	else {
	  if (have_backup) {
	    if (os) *os << "Take current step and keep current delta, since it's produced best result." << endl;
	    if (os) *os << "delta == " << delta << endl;
	    // Keep current delta because current step was arrived at through a sequence of
	    // adjustments in delta, and this value yielded best results.
	  }
	  else if (improvement == Poor) {
	    if (os) *os << "Take current step and reduce delta, because improvement is poor." << endl;
	    lambda = adjustLambda(lambda, searchflags.lambdaMin, 1.0);
	    delta  = adjustDelta(delta, lambda, searchflags.deltaMin, searchflags.deltaMax, os);
	  }
	  else if (improvement == Ok || hookstep_equals_newtonstep || (delta >= searchflags.deltaMax)) {
	    if (os) *os << "Take current step and leave delta unchanged, for the following reasons:" << endl;
	    if (os) *os << "improvement            : " << improvement << endl;
	    if (os) *os << "|newtonstep| <= delta  : " << (hookstep_equals_newtonstep ? "true" : "false") << endl;
	    if (os) *os << "delta >= deltaMax      : " << (delta >= searchflags.deltaMax    ? "true" : "false") << endl;
	    //if (os) *os << "delta has decreased    : " << (delta_has_decreased ? "true" : "false") << endl;
	    if (os) *os << "delta == " << delta << endl;
	  }
	  else {  // improvement == Good, Accurate, or NegaCurve and no restriction on increasing apply
	    if (os) *os << "Take step and increase delta, if there's room." << endl;
	    lambda = adjustLambda(lambda, 1.0, searchflags.lambdaMax);
	    delta  = adjustDelta(delta, lambda, searchflags.deltaMin, searchflags.deltaMax, os);
	  }
	}
      }
      //if (os) *os << "Recompute hookstep ? " << (recompute_hookstep ? "yes" : "no") << endl;

      if (recompute_hookstep)
	continue;
      else
	break;
    }

    if (os) *os << "Taking best step and continuing Newton iteration." <<endl;
    u += du;
    T += dT;
    sigma *= dsigma;
    dunorm = L2Norm(du);
    unorm  = L2Norm(u);

    Gx = GH;

    if (os) {
      *os << "L2Norm(G(u)) == " << L2Norm(Gx) << endl;
      *os << "rx          == " << rx << endl;
      *os << "L2Norm(dxN)   == " << L2Norm(dxN) << endl;
      *os << "L2Norm(dxH)   == " << L2Norm(dxH) << endl;
      *os << "L2Norm(du)   == " << dunorm << endl;
      *os << "       dT    == " << dT << endl;
      *os << "L2IP(du,edudx)/L2Norm(du) == " << L2IP(du,edudx)/dunorm << endl;
      *os << "L2IP(du,edudz)/L2Norm(du) == " << L2IP(du,edudz)/dunorm << endl;
      *os << "L2IP(du,edudt)/L2Norm(du) == " << L2IP(du,edudt)/dunorm << endl;
      *os << "align(dxN, prev_dxN) == " << align(dxN, prev_dxN) << endl;
      *os << "align(dxH, prev_dxH) == " << align(dxH, prev_dxH) << endl;
      *os << "align(dxH, dxN)      == " << align(dxH, dxN) << endl;
    }
    //prev_dxG = dxG;
    prev_dxH = dxH;
    prev_dxN = dxN;

    //du.save("duH" + i2s(newtonStep));
  }
  return gx;
}

// f^T(u) = time-T DNS integration of u
void f(const FlowField& u, Real T, FlowField& f_u, Real Reynolds,
       const DNSFlags& flags, const TimeStep& dtarg, int& fcount, Real& CFL, ostream* os) {

  //if (os) *os << "entering f..." << endl;

  if (T<0) {
    if (os) *os << "f: negative integration time T == " << T << endl
		<< "returning f(u,T) == (1+abs(T))*u" << endl;
    f_u = u;
    f_u *= 1+abs(T);
    //if (os) *os << "exiting f..." << endl;
    return;
  }
  if (!isfinite(L2Norm(u))) {
    if (os) *os << "L2Norm(u) == " << L2Norm(u) << endl;
    if (os) *os << "error in f: u is not finite. exiting." << endl;
    exit(1);
  }

  // Special case #1: no time steps
  if (T == 0) {
    if (os) *os << "f: T==0, no integration, returning u " << endl;
    f_u = u;
    //if (os) *os << "exiting f..." << endl;
    return;
  }

  TimeStep dt(dtarg);
  const Real nu = 1.0/Reynolds;
  FlowField p(u.Nx(), u.Ny(), u.Nz(), 1, u.Lx(), u.Lz(), u.a(), u.b());
  f_u = u;
  dt.adjust_for_T(T, false);

  // Adjust dt for CFL if necessary
  DNS dns(f_u, nu, dt, flags);
  if (dt.variable()) {
    dns.advance(f_u, p, 1);
    dt.adjust(dns.CFL(), false);
    dns.reset_dt(dt);
    f_u = u;
  }


  //if (os) *os << "\ndt == " << dt << endl;
  //if (os) *os << ", CFL == " << dns.CFL() << endl;
  //if (os) *os << "t == " << flush;

  //  t == current time in integration
  //  T == total integration time
  // dT == CFL-check interval
  // dt == DNS time-step
  //  N == T/dT,  n == dT/dt;
  //  T == N dT, dT == n dt
  //  t == s dT (s is loop index)
  //if (os) *os << "f : 0" << flush;
  for (int s=1; s<=dt.N(); ++s) {
    dns.advance(f_u, p, dt.n());

    Real t = s*dt.dT();
    CFL = dns.CFL();

    if (os) {
      if (s % 10 == 0)
	*os << iround(t) << flush;
      else if (s % 2 == 0) {
	if (CFL < dt.CFLmin())
	  *os << '<' << flush;
	else if (CFL > dt.CFLmax())
	  *os << '>' << flush;
	else
	  *os << '.' << flush;
      }
    }
    if (dt.adjust(CFL, false))
      dns.reset_dt(dt);
  }
  ++fcount;
  //if (os) *os << "finished integration..." << endl;

  if (!isfinite(L2Norm(f_u))) {
    if (os) *os << "L2Norm(f^t(u)) == " << L2Norm(f_u) << endl;
    if (os) *os << "error in f: f^t(u) is not finite. exiting." << endl;
    exit(1);
  }
  //if (os) *os << "exiting f..." << endl;

  return;
}

/// f^{N dt}(u) = time-(N dt) DNS integration of u
void f(const FlowField& u, int N, Real dt, FlowField& f_u, Real Reynolds,
       const DNSFlags& flags, ostream* os) {

  if (os) *os << "f(u, N, dt, f_u, Reynolds, flags, dt) : " << flush;
  const Real nu = 1.0/Reynolds;
  FlowField p;
  f_u = u;
  DNS dns(f_u, nu, dt, flags);
  dns.advance(f_u, p, N);
  return;
}

// G(x) = G(u,sigma,T) = (sigma f^T(u) - u) for orbits and (sigma f^T(u) - u)/T for eqbs
void G(const FlowField& u, Real T, const FieldSymmetry& sigma, FlowField& Gx,
       Real Reynolds, const DNSFlags& flags, const TimeStep& dt, bool Tnormalize, int& fcount, Real& CFL,
       ostream* os) {
  //if (os) *os << "entering G..." << endl;
  f(u, T, Gx, Reynolds, flags, dt, fcount, CFL, os);
  Gx *= sigma;
  Gx -= u;
  if (Tnormalize)
    Gx *= 1.0/T;
  //if (os) *os << "exiting G..." << endl;
  return;
}

void DG(const FlowField& u, const FlowField& du, Real T, Real dT,
	const FieldSymmetry& sigma,  const FieldSymmetry& dsigmaArg,
	const FlowField& Gx, FlowField& DG_dx,
	Real Reynolds, const DNSFlags& flags, const TimeStep& dt, bool Tnormalize,
	Real epsDu, bool centdiff, int& fcount, Real& CFL, ostream* os) {

  FlowField u_du;
  FieldSymmetry dsigma;
  FieldSymmetry sigma_dsigma;
  Real T_dT;

  Real step_magn = sqrt(L2Norm2(du) + square(dT/T) + square(dsigma.ax()) + square(dsigma.az()));
  Real eps = step_magn < epsDu ? 1 : epsDu/step_magn;

  if (centdiff) {
    Real eps2 = 0.5*eps;
    u_du  = du;
    u_du *= eps2;
    u_du += u;
    T_dT = T + eps2*dT;
    dsigma = dsigmaArg;
    dsigma *= eps2;
    sigma_dsigma = dsigma*sigma;

    G(u_du, T_dT, sigma_dsigma, DG_dx, Reynolds, flags, dt, Tnormalize, fcount, CFL, os);
    u_du  = du;
    u_du *= -eps2;
    u_du += u;
    T_dT = T - eps2*dT;
    dsigma = dsigmaArg;
    dsigma *= -eps2;
    sigma_dsigma = dsigma*sigma;
    FlowField tmp;
    G(u_du, T_dT, sigma_dsigma, tmp, Reynolds, flags, dt, Tnormalize, fcount, CFL, os);

    DG_dx -= tmp;
    DG_dx *= 1.0/eps;
  }
  else {
    u_du  = du;
    u_du *= eps;
    u_du += u;
    T_dT = T + eps*dT;
    dsigma = dsigmaArg;
    dsigma *= eps;
    sigma_dsigma = dsigma*sigma;

    G(u_du, T_dT, sigma_dsigma, DG_dx, Reynolds, flags, dt, Tnormalize, fcount, CFL,os);
    DG_dx -= Gx;
    DG_dx *= 1.0/eps;
  }
  return;
  //if (Tnormalize)
  //DG_dx *= 1.0/T;
}


GMRESHookstepFlags::GMRESHookstepFlags(SolutionType solntype_,
				       Real epsSearch_,
				       Real epsKrylov_,
				       Real epsDu_,
				       Real epsDt_,
				       Real epsGMRES_,
				       Real epsGMRESf_,
				       bool centdiff_,
				       int  Nnewton_,
				       int  Ngmres_,
				       int  Nhook_,
				       Real delta_,
				       Real deltaMin_,
				       Real deltaMax_,
				       Real deltaFuzz_,
				       Real lambdaMin_,
				       Real lambdaMax_,
				       Real lambdaRequiredReduction_,
				       Real improvReq_,
				       Real improvOk_,
				       Real improvGood_,
				       Real improvAcc_,
				       bool dudtortho_,
				       bool dudxortho_,
				       Real orthoscale_,
				       Real TscaleRel_,
				       Real Rscale_,
				       bool saveall_,
				       string outdir_,
				       ostream* logstream_)
  :
  solntype(solntype_),
  epsSearch(epsSearch_),
  epsKrylov(epsKrylov_),
  epsDu(epsDu_),
  epsDt(epsDt_),
  epsGMRES(epsGMRES_),
  epsGMRESf(epsGMRESf_),
  centdiff(centdiff_),
  Nnewton(Nnewton_),
  Ngmres(Ngmres_),
  Nhook(Nhook_),
  delta(delta_),
  deltaMin(deltaMin_),
  deltaMax(deltaMax_),
  deltaFuzz(deltaFuzz_),
  lambdaMin(lambdaMin_),
  lambdaMax(lambdaMax_),
  lambdaRequiredReduction(lambdaRequiredReduction_),
  improvReq(improvReq_),
  improvOk(improvOk_),
  improvGood(improvGood_),
  improvAcc(improvAcc_),
  dudtortho(dudtortho_),
  dudxortho(dudxortho_),
  orthoscale(orthoscale_),
  TscaleRel(TscaleRel_),
  Rscale(Rscale_),
  saveall(saveall_),
  outdir(outdir_),
  logstream(logstream_)
{}

bool GMRESHookstepFlags::relative() const {
  return (solntype == TravelingWave || solntype == RelativePeriodicOrbit);
}
bool GMRESHookstepFlags::orbit() const {
  return (solntype == PeriodicOrbit || solntype == RelativePeriodicOrbit);
}
// The following is meant as a guide to the structure and notation
// of the above code. See Divakar Viswanath, "Recurrent motions
// within plane Couette turbulence", J. Fluid Mech. 580 (2007),
// http://arxiv.org/abs/physics/0604062 for a higher-level presentation
// of the algorithm.
//
// I've biased the notation towards conventions from fluids and from
// Predrag Cvitanovic's ChaosBook, and away from Dennis & Schnabel and
// Trefethen. That is, f stands for the finite-time map of plane Couette
// dynamics, not the residual being minimized.
//
// Note: in code, dt stands for the integration timestep and
//                dT stands for Newton-step increment of period T
// But for beauty's sake I'm using t and dt for the period and
// its Newton-step increment in the follwoing comments.
//
// Let
//     f^t(u) be the time-t map of Navier-Stokes computed by DNS.
//     sigma    be a symmetry of the flow
//
// We seek solutions of the equation
//
//    G(u,sigma,t) = sigma f^t(u) - u = 0                       (1)
//
// t can be a fixed or a free variable, as can sigma. For plane
// Couette the potential variables in sigma are ax,ax in the
// phase shifts x -> x + ax Lx and z -> z + az Lz.
//
// Let x be the vector of unknown real numbers being sought.
// For simplicity's sake in this exposition I will assume that
// we're looking for a relative periodic orbit, so that the
// period t of the orbit and the phase shifts ax and az are
// unknowns. Then x = (u, sigma, t). Of course by this I really
// that x is the finite set of real-valued variables that
// parameterize the discrete representation of u (e.g. the
// real and imaginary parts of the spectral coefficients),
// together with the real-valued variables that parameterize
// sigma, and the real number t. To be perfectly pedantic (and
// because it will come in handy later), let there be M
// independent real numbers {u_m} in the discrete representation
// of u, and let
//
//    {u_m} = P(u),  u_m = P_m(u)                           (2)
//
// represent the map between continuous fields u and the discrete
// representation {u_m}. Then x = (u, sigma, t) is an M+3 dimensional
// vector
//
//    x_m = (u_1, u_2, ..., u_M, ax, az, t)
//
// and the discrete equation to be solved is
//
//   G(x) = 0                                               (3)
//
// where G is a vector-valued function of dimension M.
//
//   G_m(x)   = P_m(sigma f^t(u) - u)                         (4)
//
// At this point have three fewer equations in (4) than unknowns
// in x because the equivariance of the flow means that solutions
// are indeterminate under time and phase shifts.
//
// The Newton algorithm for solving G(x) = 0 is
//
// Let x* = (u*,sigma*,t*) be the solution, i.e. G(x*) = G(u*,sigma*, t*) = 0.
// Suppose we start with an initial guess x near x*.
// Let
//     x* = x + dxN       (dxN is the Newton step)           (5)
//     u* = u + duN
// sigma* = sigma + dsigmaN
//     t* = t + dtN
//
// Then G(x*) = G(x + dxN)
//          0 = G(x) + DG(x) dxN + O(|dxN|^2)
//
// Drop h.o.t and solve the Newton equation
//
//    DG(x) dxN = -G(x)                                     (6)
//
// Because DG is a M x (M+3) matrix, we have to supplement this
// system with three constraint equations to have a unique solution
// dxN.
//
//   (duN, du/dt) = 0                                       (7)
//   (duN, du/dx) = 0
//   (duN, du/dz) = 0
//
// ( , ) here signifies an inner product, so these are orthogonality
// constraints preventing the Newton step from going in the directions
// of the time, x, and z equivariance. That forms an (M+3) x (M+3)
// dimensional system
//
//    A dxN = b                                             (8)
//
//  where the first M rows are given by (6) and the last three by (7).
//
//
// Since since M is very large, we will use the iterative GMRES algorithm
// to find an approximate solution dxN to the (M+3)x(M+3) system A dxN = b.
// GMRES requires multiple calculations A dx for test values of dx. The
// left-hand side of (6) can be approximated with finite differencing:
//
//   DG(x) dx = 1/e (G(x + e dx) - G(x)) where e |dx| << 1.
//            = 1/e P[(sigma+dsigma) f^{t+dt}(u+du) - (u+du) - (sigma f^{t}(u) - u)]
//            = 1/e P[(sigma+dsigma) f^{t+dt}(u+du) - sigma f^{t}(u) - du]
//
// And the left-hand sides of (7) are a simple evaluation of an inner
// product. For details of the GMRES algorithm, see Trefethen and Bau.

// That gives us the Newton step dxN. However if we are too far from
// the solution x*, the linearization inherent in the Newton algorithm
// will not be accurate. At this point we switch from the pure Newton
// algorithm to a constrained minimization algorithm called the
// "hookstep", specially adapted to work with GMRES.

// Hookstep algorithm: If the Newton step dxN does not decrease
// the residual || G(x+dxN) || sufficiently, we obtain a smaller step
// dxH by minimizing || A dxH - b ||^2 subject to ||dxH||^2 <= delta^2,
// rather using the Newton step, which solves A dx = b.

// GMRES-Hookstep algorithm: Since A is very high-dimensional, we
// further constrain the hookstep dxH to lie within the Krylov subspace
// obtained from the solution of the Newton step equations. I.e.
// we minimize the norm of the projection of A dxH - b onto the Krylov
// subspace. The nth GMRES iterate gives matrices Qn, Qn1, and Hn such
// that
//
//   A Qn = Qn1 Hn                                              (9)
//
// where Qn is M x n, Qn1 is M x (n+1), and Hn is (n+1) x n.
// Projection of A dx - b on (n+1)th Krylov subspace is Qn1^T (A dx - b)
// Confine dx to nth Krylov subspace: dxH = Qn sn, where sn is n x 1.
// Then minimize
//
//   || Qn1^T (A dxH - b) || = || Qn1^T A Qn sn - Qn1^T b ||
//                          = || Qn1^T Qn1 Hn sn - Qn1^T b ||
//                          = || Hn sn - Qn1^T b ||            (10)
//
// subject to ||dxH||^2 = || sn ||^2 <= delta^2. Note that the quantity in
// the norm on the right-hand side of (10) is only n-dimensional, and n is
// small (order 10 or 100) compared to M (10^5 or 10^6).
//
// Do minimization of the RHS of (10) via SVD of Hn = U D V^T.
//   || Hn sn - Qn1^T b || = || U D V^T sn - Qn1^T b ||
//                         = || D V^T sn - U^T Qn1^T b ||
//                         = || D sh - bh ||
//
// where sh = V^T sn and bh = U^T Qn1^T b. Now we need to
// minimize || D sh - bh ||^2 subject to ||sh||^2 <= delta^2
//
// From Divakar Viswanath (personal communication):
//
// Since D is diagonal D_{i,i} = d_i, the constrained minimization problem
// can be solved easily with Lagrange multipliers. The solution is
//
//   sh_i = (bh_i d_i)/(d^2_i + mu)
//
// for the mu such that ||sh||^2 = delta^2. The value of mu can be
// found with a 1d Newton search for the zero of
//
//   Phi(mu) = ||sh(mu)||^2 - delta^2
//
// A straight Newton search a la mu -= Phi(mu)/Phi'(mu) is suboptimal
// since Phi'(mu) -> 0 as mu -> infty with Phi''(mu) > 0 everywhere, so
// we use a slightly modified update rule for the Newton search over mu.
// Please refer to Dennis and Schnabel regarding that. Then, given the
// solution sh(mu), we compute the hookstep solution s from
//
//   dxH = Qn sn = Qn V sh
//
//
// How do we know a good value for delta? Essentially, by comparing
// the reduction in residual obtained by actually taking the hookstep
// dxH to the reduction predicted by the linearized model  G(x+dxH) =
// G(x) + DG(x) dxH. If the reduction is accurate but small, we increase
// delta by small steps until the reduction is marginally accurate but larger.
// If the reduction is poor we reduce delta.
//
// (Note: the trust-region optimization is actually performed in terms of the
// squared residual, so the comments in the code refer to a quadratic rather than
// linear model of the residual.)
//
// The heuristics associated with adjusting the trust region are fairly complex.
// Dennis and Schnabel has a decent description, but it took quite a bit of effort
// to translate that description into an algorithm. Rather than attempt to
/// translate the algorithm back into prose, I recommend you read Dennis and
// Schnabel and then refer to the code and comments. This portion of findorbit.cpp
// is very liberally commented.



// References:
//
// Divakar Viswanath, "Recurrent motions within plane Couette turbulence",
// J. Fluid Mech. 580 (2007), http://arxiv.org/abs/physics/0604062
//
// Lloyd N. Trefethen and David Bau, "Numerical Linear Algebra", SIAM,
// Philadelphia, 1997.
//
// Dennis and Schnabel, "Numerical Methods for Unconstrained Optimization
// and Nonlinear Equations", Prentice-Hall Series in Computational Mathematics,
// Englewood Cliffs, New Jersey, 1983.
//

//#endif // HAVE_OCTAVE
// End of Octave-dependent code
// ======================================================================


} // namespace channelflow
