#ifndef CHANNELFLOW_OCTUTILS_H
#define CHANNELFLOW_OCTUTILS_H

#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iomanip>
#include <string>
#include <iostream>
#include <vector>

#include <octave/oct.h>
#include <octave/dColVector.h>
#include <octave/dMatrix.h>
#include <octave/EIG.h>

#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/quicksort.h"
#include "channelflow/utilfuncs.h"

namespace channelflow {


// ====================================================================
// All Octave-dependent code goes within this ifdef

void print(const Matrix& x);
void print(const ColumnVector& x);
void save(const Matrix& A, const std::string& filebase);
void save(const ColumnVector& x, const  std::string& filebase);
void save(const ComplexColumnVector& x, const std::string& filebase);
void setToZero(Matrix& A);
void setToZero(ColumnVector& x);

Real L2Norm(const ColumnVector& x);
Real L2Dist(const ColumnVector& x,  const ColumnVector& y);
Real L2Norm2(const ColumnVector& x);
Real L2Dist2(const ColumnVector& x,  const ColumnVector& y);
Real L2IP(const ColumnVector& u, const ColumnVector& v);

void operator*=(ColumnVector& x, Real c);

Real L2Norm(const ComplexColumnVector& x);
Real L2Dist(const ComplexColumnVector& x,  const ComplexColumnVector& y);
Real L2Norm2(const ComplexColumnVector& x);
Real L2Dist2(const ComplexColumnVector& x,  const ComplexColumnVector& y);
void operator*=(ComplexColumnVector& x, Complex c);

enum BasisType {L2Basis, AdHocBasis};

Real V2Norm2(const FlowField& u); // L2Norm2(field2vector(u));
Real V2Norm(const FlowField& u);
Real V2IP(const FlowField& u, const FlowField& v);

Real Norm2(const FlowField& u, BasisType norm);
Real Norm(const FlowField& u, BasisType norm);
Real IP(const FlowField& u, const FlowField& v, BasisType norm);


void quicksort(ComplexColumnVector& z); // sort by norm, decreasing
void quicksort(ComplexColumnVector& Lambda, ComplexMatrix& V); // ditto

// The field2vector and vector2field functions assume zero divergece and no-slip BCs.
void field2vector(const FlowField& u, ColumnVector& v,
		  const std::vector<RealProfileNG>& basis);
void vector2field(const ColumnVector& v, FlowField& u,
		  const std::vector<RealProfileNG>& basis);

void field2vector(const std::vector<BasisFunc>& basis,
		  const FlowField& u, ComplexColumnVector& v);
void vector2field(const std::vector<BasisFunc>& basis,
		  const ComplexColumnVector& v, FlowField& u);

void field2vector(const FlowField& u, ColumnVector& v);

void vector2field(const ColumnVector& v, FlowField& u);

void fixdivnoslip(FlowField& u);

void leastsquares(const Matrix& A, ColumnVector& b, ColumnVector& x,
		  Real& residual);

void solve(const Matrix& Ut, const DiagMatrix& D, const Matrix& C,
	   const ColumnVector& b, ColumnVector& x);

// f^T(u) = time-T DNS integration of u
void f(const FlowField& u, Real T, FlowField& f_u, Real Reynolds,
       const DNSFlags& flags, const TimeStep& dt, int& fcount, Real& CFL, std::ostream* os=0);

// f^(n.dt)(u) == time-(n.dt) DNS integration of u
void f(const FlowField& u, int n, Real dt, FlowField& f_u, Real Reynolds,
       const DNSFlags& flags, std::ostream* os=0);

// G(u,sigma,T) = (sigma f^T(u) - u)   for orbits and
//                (sigma f^T(u) - u)/T for eqbs, tws
void G(const FlowField& u, Real T, const FieldSymmetry& sigma, FlowField& G_u,
       Real Reynolds, const DNSFlags& flags, const TimeStep& dt, bool Tnormalize,
       int& fcount, Real& CFL, std::ostream* os=0);

void DG(const FlowField& u, const FlowField& du, Real T, Real dT,
	const FieldSymmetry& sigma,  const FieldSymmetry& dsigma,
	const FlowField& Gx, FlowField& DG_dx, Real Reynolds, const DNSFlags& flags,
	const TimeStep& dt, bool Tnormalize, Real eps, bool centerdiff, int& fcount,
	Real& CFL, std::ostream* os=0);




// Arnoldi iteration to estimate eigenvalues of matrix A. Usage:
// Arnoldi arnoldi(N, b);
// for (int n=0; n<N; ++n) {
//   ColumnVector q  = arnoldi.testVector();
//   ColumnVector Aq = A*q; // or however else you calculate A*q
//   arnoldi.iterate(Aq);
//   ComplexColumnVector ew = arnoldi.ew(); // current estimate of eigenvalues
// }

class Arnoldi {

public:
  Arnoldi();
  Arnoldi(const ColumnVector& b, int Niterations, Real minCondition=1e-13);

  const ColumnVector& testVector() const; // get test vector q
  void iterate(const ColumnVector& Aq);   // tell Arnoldi the value of Aq

  int n() const;     // current iteration number
  int Niter() const; // total number iterations

  const ComplexColumnVector& ew(); // current estimate of eigenvals
  const ComplexMatrix& ev();       // current estimate of eigenvecs

private:
  int M_;
  int Niter_;
  int n_;     // current iteration number
  Real condition_;

  Matrix H_;
  Matrix Q_;
  Matrix Vn_;
  ColumnVector qn_;
  ColumnVector v_;

  ComplexColumnVector ew_;
  ComplexMatrix ev_;

  void eigencalc();
};


// Iterative GMRES solution to Ax=b. Usage:
// GMRES gmres(b, N);
// for (int n=0; n<N; ++n) {
//   ColumnVector q  = gmres.testVector();
//   ColumnVector Aq = A*q; // or however else you calculate A*q
//   gmres.iterate(Aq);
//   ColumnVector x = gmres.solution(); // current estimate of soln
//   cout << "krylov residual == " << gmres.residual() << endl;
// }

// Iterative solution to Ax=b. Usage:
class GMRES {
public:
  GMRES();
  GMRES(const ColumnVector& b, int Niterations, Real minCondition=1e-13);

  const ColumnVector& testVector() const;
  void iterate(const ColumnVector& A_testvec);

  const ColumnVector& solution() const;
  const ColumnVector& guess() const;
  Real residual() const;  // |Ax-b|/|b|

  int n() const;
  int Niter() const;

  Matrix Hn() const;  // Hn  = (n+1) x n submatrix of H
  Matrix Qn() const;  // Qn  = M x n     submatrix of Q
  Matrix Qn1() const; // Qn1 = M x (n+1) submatrix of Q

private:
  int M_;     // dimension of vector space
  int Niter_; // max number of iterations
  int n_;     // current iteration number
  Real condition_;

  Matrix H_;
  Matrix Q_;
  Matrix Vn_;
  ColumnVector x0_; // guess;
  ColumnVector v_;
  ColumnVector qn_;
  ColumnVector xn_;
  Real bnorm_;
  Real residual_;
};

// Iterative solution to Ax=b with initial guess x0.
// Internally, let A (x0+y) = b and solve A y = b - A x0 with GMRES
// Residual is |Ax-b|/|b|, not |Ax-b|/|b-Ax0|
class GMRESguess {
public:
  GMRESguess(int Niterations, const ColumnVector& b, Real minCondition=1e-10);
  GMRESguess(int Niterations, const ColumnVector& b, const ColumnVector& x0, const ColumnVector& Ax0,
	     Real minCondition=1e-13);

  const ColumnVector& testVector() const;
  void iterate(const ColumnVector& A_testvec);

  ColumnVector solution() const;
  Real krylovResidual() const;

  int n() const;
  int Niter() const;

private:
  GMRES gmres_;
  ColumnVector x0_;
  Real residualScale_; // |b-Ax0|/|b|
};



class HookstepParams {
public:
  HookstepParams(Real epsSearch = 1e-14,
		 Real epsDx     = 1e-7,
		 bool centdiff  = false,
		 int  Nnewton   = 20,
		 int  Nhook     = 20,
		 Real deltaInit = 0.01,
		 Real deltaMin  = 1e-12,
		 Real deltaMax  = 1e+12,
		 Real deltaFuzz = 0.01,
		 Real lambdaMin = 0.10,
		 Real lambdaMax = 2.0,
		 Real lambdaRequiredReduction = 0.5,
		 Real improvReq = 1e-03,
		 Real improvOk   = 0.10,
		 Real improvGood = 0.75,
		 Real improvAcc  = 0.10,
		 bool pausing    = false,
		 std::ostream* logstream= &std::cout);

  Real epsSearch; // stop search if |f(x)| < epsSearch
  Real epsDx;     // scale for perturbs in finite-diff estimate of Df
  bool centdiff;  // used centered differences in estimating Df
  int  Nnewton;   // max number of Newton step
  int  Nhook;     // max number of hookstep iterations per Newton
  Real deltaInit; // initial trust region radius
  Real deltaMin;  // stop if radius of trust region gets this small
  Real deltaMax;  // don't let trust region radius get bigger than this
  Real deltaFuzz; // accept steps within (1+/-deltaFuzz)*delta
  Real lambdaMin; // minimum delta shrink rate
  Real lambdaMax; // maximum delta expansion rate
  Real lambdaRequiredReduction; // when reducing delta, reduce by at least this factor.
  Real improvReq; // reduce delta and recompute hookstep if residual is worse than this fraction of what we'd expect from gradient
  Real improvOk;  // accept step and keep same delta if improvement is better than this fraction of quadratic mode
  Real improvGood;// accept step and increase delta if improvement is better than this fraction of quadratic model
  Real improvAcc; // recompute hookstep with larger trust region if improvement is within this fraction quadratic prediction.
  ColumnVector xscale;     // rescale x during hookstep search, x -> x./xscale. Default == don't
  ColumnVector fscale;     // rescale f during hookstep search, f -> f./fscale. Default == don't
  bool pausing;
  std::ostream* logstream; // write search log here
  //string logfile;
};


enum HookstepPhase {ConstantDelta, ReducingDelta, IncreasingDelta, Finished};
enum ResidualImprovement {Unacceptable, Poor, Ok, Good, Accurate, NegaCurve};

std::ostream& operator<<(std::ostream& os, HookstepPhase p);
std::ostream& operator<<(std::ostream& os, ResidualImprovement i);

Real adjustLambda(Real lambda, Real lambdaMin, Real lambdaMax, std::ostream* os=0);
Real adjustDelta(Real delta, Real rescale, Real deltaMin, Real deltaMax, std::ostream* os=0);


// This enum is used for labeling diagnostic output data
enum fEvalType {fEval, DfEval, HookstepEval};

typedef ColumnVector (*Rn2Rnfunc)(const ColumnVector& x, int label, bool cacheable);

void rescale(ColumnVector& x, const ColumnVector& xscale); // x -> x./xscale
void unscale(ColumnVector& x, const ColumnVector& xscale); // x -> x.*xscale

ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0);
ColumnVector hookstepSearch(Rn2Rnfunc f, const ColumnVector& x0, const HookstepParams& p, int& nsteps);

// returns u dot v/(|u||v|) or zero if either has zero norm
Real align(const ColumnVector& u, const ColumnVector& v);

typedef Real (*R2Rfunc)(Real x);

//Real bisectSearch(R2Rfunc f, Real a, Real b, Real feps=1e-14, int maxsteps=50);

//enum HookstepPhase {ConstantDelta, ReducingDelta, IncreasingDelta, Finished};
//std::ostream& operator<<(std::ostream& os, HookstepPhase p);

//enum ResidualImprovement {Unacceptable, Poor, Ok, Good, Accurate, NegaCurve};
//std::ostream& operator<<(std::ostream& os, ResidualImprovement i);

//Real adjustLambda(Real lambda, Real lambdaMin, Real lambdaMax);
//Real adjustDelta(Real delta, Real rescale, Real deltaMin, Real deltaMax);

enum SolutionType {Equilibrium, TravelingWave, PeriodicOrbit, RelativePeriodicOrbit};

class GMRESHookstepFlags {
public:
  SolutionType solntype;

  Real epsSearch; // stop search if L2Norm(s f^T(u) - u) < epsEQB
  Real epsKrylov; // min. condition # of Krylov vectors
  Real epsDu;     // relative size of du to u in linearization
  Real epsDt;     // size of dT in linearization of f^T about T;
  Real epsGMRES;  // stop GMRES iteration when Ax=b residual is < this;
  Real epsGMRESf; // accept final GMRES iterate if residual is < this);
  bool centdiff;  // centered differencing to estimate differentials

  int  Nnewton;   // max number of Newton steps
  int  Ngmres;    // max number of GMRES iterations per restart;
  int  Nhook;     // max number of hookstep iterations per Newton

  Real delta;      // initial radius of trust region
  Real deltaMin;   // stop if radius of trust region gets this small
  Real deltaMax;   // maximum radius of trust region
  Real deltaFuzz;  // accept steps within (1+/-deltaFuzz)*delta
  Real lambdaMin;  // minimum delta shrink rate
  Real lambdaMax;  // maximum delta expansion rate
  Real lambdaRequiredReduction; // when reducing delta, reduce by at least this factor.
  Real improvReq;  // reduce delta and recompute hookstep if improvement is worse than this fraction of what we'd expect from gradient");
  Real improvOk ;  // accept step and keep same delta if improvement is better than this fraction of quadratic model
  Real improvGood; // accept step and increase delta if improvement is better than this fraction of quadratic model
  Real improvAcc;  // recompute hookstep with larger trust region if improvement is within this fraction quadratic prediction.

  bool dudtortho;  // require orthogonality of step to dudt
  bool dudxortho;  // require orthogonality of step to dudx and dudz");
  Real orthoscale; // rescale orthogonality constraint
  Real TscaleRel;  // scale time in hookstep: |T| = Ts*|u|/L2Norm(du/dt)
  Real Rscale;     // scale relative-search variables by this factor;
  bool saveall;    // save fields at every newton step
  std::string outdir;  // directory for storing solutions
  std::ostream* logstream; // output stream for logging solution algorithm

  GMRESHookstepFlags(SolutionType solntype = Equilibrium,
		     Real epsSearch = 1e-13,
		     Real epsKrylov = 1e-14,
		     Real epsDu = 1e-07,
		     Real epsDt = 1e-07,
		     Real epsGMRES = 1e-03,
		     Real epsGMRESf = 0.05,
		     bool centdiff = false,
		     int  Nnewton = 30,
		     int  Ngmres = 50,
		     int  Nhook = 20,
		     Real delta = 1e-02,
		     Real deltaMin = 1e-08,
		     Real deltaMax = 1e-01,
		     Real deltaFuzz = 1e-02,
		     Real lambdaMin = 0.2,
		     Real lambdaMax = 1.5,
		     Real lambdaRequiredReduction = 0.5,
		     Real improvReq = 1e-03,
		     Real improvOk  = 1e-01,
		     Real improvGood = 0.75,
		     Real improvAcc = 1e-01,
		     bool dudtortho = true,
		     bool dudxortho = true,
		     Real orthoscale = 1,
		     Real TscaleRel = 20,
		     Real Rscale = 1.0,
		     bool saveall = false,
		     std::string outdir = "./",
		     std::ostream* logstream = &std::cout);

  //GMRESHookstepFlags(ArgList& args);

  bool relative() const;
  bool orbit() const;
};

// Find solution of equation sigma f^T(u) - u == 0. Inputs are initial
// guess and return as solution values. Real return value is residual
// L2Norm(sigma f^T(u) - u)
Real GMRESHookstep(FlowField& u, Real& T, FieldSymmetry& sigma, Real Reynolds,
		   const GMRESHookstepFlags& searchflags, const DNSFlags& dnsflags,
		   const TimeStep& dt);

} // namespace channelflow

#endif // OCTAVE_UTILITIES_H
