#include <iostream>
#include <fstream>
#include <iomanip>
#include "channelflow/flowfield.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/utilfuncs.h"

using namespace std;
using namespace channelflow;

int main(int argc, char* argv[]) {

  string purpose("Construct a field with random field with zero divergence and Dirichlet BCs\n"
		 "The field is u = sum_{jkl} a_{jkl} T_l(y) exp(2 pi i (jx/Lx + kz/Lz)) where\n"
		 "a_{jkl} = (random # in [-1 1]) * smoothness^{|l| + |j| + |k|}, with corrections\n"
		 "to assure u has zero divergence and no-slip BCs, and with rescaling\n"
		 "u = magnitude/L2Norm(u)");

  ArgList args(argc, argv, purpose);

  const int Nx       = args.getint("-Nx", "--Nx", "# x gridpoints");
  const int Ny       = args.getint("-Ny", "--Ny", "# y gridpoints");
  const int Nz       = args.getint("-Nz", "--Nz", "# z gridpoints");
  const Real alpha   = args.getreal("-a", "--alpha", 0, "Lx = 2 pi/alpha");
  const Real gamma   = args.getreal("-g", "--gamma", 0, "Lz = 2 pi/gamma");
  const Real lx      = (alpha == 0.0) ? args.getreal("-lx", "--lx", "Lx = 2 pi lx") : 1/alpha;
  const Real lz      = (gamma == 0.0) ? args.getreal("-lz", "--lz", "Lz = 2 pi lz") : 1/gamma;
  const int  seed    = args.getint("-sd",  "--seed",       1, "seed for random number generator");
  const Real smooth  = args.getreal("-s",  "--smoothness", 0.4, "smoothness of field, 0 < s < 1");
  const Real magn    = args.getreal("-m",  "--magnitude", 0.20, "magnitude  of field, 0 < m < 1");
  const bool meanfl  = args.getflag("-mf",  "--meanflow", "perturb the mean");

  const bool s1symm  = args.getflag("-s1",  "--s1-symmetry", "satisfy s1 symmetry");
  const bool s2symm  = args.getflag("-s2",  "--s2-symmetry", "satisfy s2 symmetry");
  const bool s3symm  = args.getflag("-s3",  "--s3-symmetry", "satisfy s3 symmetry");

  const string uname = args.getstr(1, "<fieldname>", "output file");
  args.check();
  args.save("./");

  srand48(seed);

  const Real Lx = 2*pi*lx;
  const Real a = -1.0;
  const Real b =  1.0;
  const Real Lz = 2*pi*lz;

  FlowField u(Nx,Ny,Nz,3,Lx,Lz,a,b);
  u.addPerturbations(u.kxmaxDealiased(), u.kzmaxDealiased(), 1.0, 1-smooth, meanfl);

  if (s1symm)  {
    FieldSymmetry s1( 1, 1,-1, 0.5, 0.0);
    u += s1(u);
  }
  if (s2symm)  {
    FieldSymmetry s2(-1,-1, 1, 0.5, 0.5);
    u += s2(u);
  }
  if (s3symm)  {
    FieldSymmetry s3(-1,-1,-1, 0.0, 0.5);
    u += s3(u);
  }

  u *= magn/L2Norm(u);
  u.setPadded(true);
  u.save(uname);
}
