// License declaration at end of file

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <sys/stat.h>

#include "channelflow/flowfield.h"
#include "channelflow/vector.h"
#include "channelflow/chebyshev.h"
#include "channelflow/tausolver.h"
#include "channelflow/dns.h"
#include "channelflow/symmetry.h"
#include "channelflow/turbstats.h"
#include "channelflow/utilfuncs.h"

using namespace std;
using namespace channelflow;

int main(int argc, char* argv[]) {

  string purpose("produce I,D,E, time series and means from u(x,t) time series");
  ArgList args(argc, argv, purpose);

  const Real T0        = args.getreal("-T0", "--T0", 0,   "start time");
  const Real T1        = args.getreal("-T1", "--T1", 500, "end time");
  const Real dT        = args.getreal("-dT", "--dT", 1.0, "save interval");
  const string udir    = args.getpath("-d",  "--datadir", "data/", "flowfield series directory");
  const string ulabel  = args.getstr ("-ul", "--ulabel",  "u",  "flowfield filename label");
  const string odir    = args.getpath("-o",  "--outdir",  "orbprops/", "output directory");
  const bool closed    = args.getflag("-c",  "--closed",  "closed orbit, so don't count connecting point twice");
  const bool channel  = args.getflag("-ch", "--channel", "channelflow instead of plane Couette");
  const bool stddev    = args.getflag("-sd",  "--stddev",  "compute std dev <|u-<u>|^2>^(1/2)");
  const int digits     = args.getint ("-dg", "--digits",  8,  "# digits in output");
  const Real Reynolds  = args.getreal("-R", "--Reynolds", 400, "Reynolds numbers");

  const bool inttime = (abs(dT - int(dT)) < 1e-12) ? true : false;
  args.check();
  args.save("./");
  mkdir(odir);

  const char s = ' ';
  const Real nu = 1.0/Reynolds;

  FlowField umean(udir + ulabel + t2s(T0, inttime));
  umean.setToZero();

  const int Nx = umean.Nx();
  const int Ny = umean.Ny();
  const int Nz = umean.Nz();
  const Real Lx = umean.Lx();
  const Real Lz = umean.Lz();
  const Real ya = umean.a();
  const Real yb = umean.b();
  FlowField tmp(Nx,Ny,Nz,9,Lx,Lz,ya,yb);


  Vector y = chebypoints(Ny, ya, yb);
  ChebyCoeff U(Ny,ya,yb,Spectral);
  if (channel) {
    U[0] = 0.5;
    U[2] = -0.5;
  }
  else
    U[1] = 1;

  TurbStats stats(U, nu);
  ofstream os((odir + "IDEt.asc").c_str());

  os << "% ";
  for (int n=0; n<argc; ++n)
    os << argv[n] << s;
  os << endl;
  os << "% forcing(u+U) dissipation(u+U) energy(u+U) energy(u)" << endl;
  os << setprecision(lesser(digits, 17));

  Real I = 0.0;
  Real D = 0.0;
  Real E = 0.0;
  Real En = 0.0;
  int count = 0;
  os << setprecision(17);

  for (Real t=T0; t<=T1; t += dT) {

    string uname = ulabel + t2s(t, inttime);
    cout << uname << s << flush;
    FlowField u(udir + uname);
    umean += u;
    stats.addData(u,tmp);
    Real en = 0.5*L2Norm2(u);

    u += U;
    Real i = forcing(u);
    Real d = dissipation(u);
    Real e = 0.5*L2Norm2(u);
    os << i <<s<< d <<s<< e <<s<< en << endl;

    // If orbit is closed, skip first value so as not to count it twice
    if (closed && t==T0)
      continue;

    I += i;
    D += d;
    E += e;
    En += en;
    ++count;
  }
  ofstream os2((odir + "IDE.asc").c_str());
  os2 << "% <forcing(u+U)> <dissipation(u+U)> <energy(u+U)> <energy(u)>" << endl;
  os2 << I/count <<s<< D/count <<s<< E/count <<s<< En/count<< endl;

  umean *= 1.0/count;
  umean.save(odir + "umean");

  stats.msave(odir + "stats", false);
  stats.msave(odir + "wallstats", true);

  if (stddev) {
    cout << "Computing std dev ..." << flush;
    Real sum = 0.0;
    count = 0;
    for (Real t=T0; t<=T1; t += dT) {
      string uname = ulabel + t2s(t, inttime);
      cout << uname << s << flush;
      FlowField u(udir + uname);
      u -= umean;
      sum += L2Norm2(u);
      ++count;
    }
    cout << endl;
    cout << "std dev == " << sqrt(sum/count) << endl;
  }
  cout << "done!" << endl;
}

/* project.cpp: project time series of fields onto a given basis
 * channelflow-1.1 PCF-utils
 *
 * Copyright (C) 2001-2007  John F. Gibson
 *
 * gibson@cns.physics.gatech.edu  jfg@member.fsf.org
 *
 * Center for Nonlinear Science
 * School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 * 404 385 2509
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation version 2
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */
